// Copyright 1992 by Jon Dart. All Rights Reserved.
// $Revision: 1.2 $
//
#ifndef _MOVE_GENERATOR_H
#define _MOVE_GENERATOR_H

#include "board.h"
#include "constant.h"

class Move_Generator
{
    enum Phase {HASH_MOVE_PHASE, WINNING_CAPTURE_PHASE,
      KILLER1_PHASE, KILLER2_PHASE, HISTORY_PHASE,
      HISTORY_PHASE2, LOSERS_PHASE, LAST_PHASE };

    public:
		
    enum {MaxMoves = 200};		
		
    Move_Generator( const Board &ABoard, const unsigned ply, const Move pvMove);
	
    // Generate the next move, in sorted order, NullMove if none left
    Move NextMove() {
       if (index >= batch_count) {
          if ((batch_count = getBatch(batch,index)) == 0) 
             return NullMove;
       }
       return batch[index++];
    }

    // Generate the next check evasion, NullMove if none left
    Move NextEvasion();
        
    // Generate only non-capturing moves.
    int GenerateNonCaptures(Move *moves);

    // fill array "moves" with moves, starting at "index", returns
    // # of moves (0 if no more). This tries to generate moves in stages,
    // so it should be called repeatedly until the return value is zero.
    // 		
    // The moves returned are generally "pseudo-legal", i.e. they may
    // involve moves into check.  However, if the side to move is in
    // check, then all moves returned are strictly legal.		
    int GenerateAllMoves(Move *moves,int repeatable);

    // Like Generate_Moves, but only generates captures, promotions,
    // and check evasions.
    int Generate_Captures( Move *moves );
		
    // returns a list of the moves that may be used to escape
    // check.  Unlike the regular move generation routine, all
    // moves are checked for legality.
    unsigned Check_Evasions(Move * moves, int captures_only);

    unsigned Moves_Generated()
    {
       return moves_generated;
    }
		
    static void clear_killer();
    // clear out the killer move array.
    
    static void update_killers();
		
    // clear out the history arrays.
    static void clear_history();
    
    static void set_killer(const Board& board, const Move & move, 
                          const unsigned ply, const unsigned depth);

    static void update_history( const Board &board,
      const Move &move, const unsigned depth );

    int num_evasions();

	private:
		
    int getBatch(Move *&batch,int &index);
    int OrderWinningCaptures (const Board &board, Move *moves, int index,
      int captures, Move *losers, int &losers_count, int ply);

    const Board &board;
    int my_ply;
    int moves_generated;
    int phase, losers_count,index,batch_count,evasion_count;
    Move m_hashMove;
    int non_captures_count,history_count;
    Move *batch;
    Move losers[100];
    Move moves[MaxMoves];
        
    inline void SetMove( const Square source, const Square dest,
       const PieceType promotion,
       Move *moves, unsigned &NumMoves);

};

#endif
