// Copyright 1994-2004 by Jon Dart.  All Rights Reserved.

#include "scoring.h"
#include "bearing.h"
#include "constant.h"
#include "util.h"
#include "pawnentr.h"
#include "kingpawn.h"
#include "bhash.h"
#include "config.h"
#include "debug.h"
#include "bhash.h"
extern "C" {
#include <stdio.h>
};

#include <memory.h>

//#define EVAL_DEBUG
//#define PAWN_DEBUG
//#define LAZY_DEBUG

static const int KnightScoresB[] =
{-11, -7, -7, -7, -7, -7, -7, -11,
   -4, 0, 1, 1, 1, 1, 0, -4,
   -4, 1, 3, 2, 2, 3, 1, -4,
   -4, 1, 2, 7, 7, 2, 1, -4,
   -4, 3, 2, 7, 7, 2, 3, -4,
   -4, 3, 4, 4, 4, 4, 3, -4,
   -4, 0, 1, 1, 1, 1, 0, -4,
   -4, 0, 0, 0, 0, 0, 0, -4};

static const int KnightScoresW[] =
{-4, 0, 0, 0, 0, 0, 0, -4,
   -4, 0, 1, 1, 1, 1, 0, -4,
   -4, 3, 4, 4, 4, 4, 3, -4,
   -4, 3, 2, 7, 7, 2, 3, -4,
   -4, 1, 2, 7, 7, 2, 1, -4,
   -4, 1, 3, 2, 2, 3, 1, -4,
   -4, 0, 1, 1, 1, 1, 0, -4,
   -11, -7, -7, -7, -7, -7, -7, -11};

static const int BishopCenterScores[] =
{-2, -2, -2, -2, -2, -2, -2, -2,
   -2, 0, 0, 0, 0, 0, 0, -2,
   -2, 0, 1, 1, 1, 1, 0, -2,
   -2, 0, 1, 2, 2, 1, 0, -2,
   -2, 0, 1, 2, 2, 1, 0, -2,
   -2, 0, 1, 1, 1, 1, 0, -2,
   -2, 0, 0, 0, 0, 0, 0, -2,
   -2, -2, -2, -2, -2, -2, -2, -2};

static const int BishopPositionB[] =
{-3, -6, -6, -6, -6, -6, -6, -3,  
   0, 3, 0, 0, 0, 0, 3, 0,
   0, 0, 3, 0, 0, 3, 0, 0,
   0, 0, 0, 3, 3, 0, 0, 0,
   0, 0, 0, 3, 3, 0, 0, 0,
   0, 0, 3, 0, 0, 3, 0, 0,
   0, 3, 0, 0, 0, 0, 3, 0,
   3, 0, 0, 0, 0, 0, 0, 3};

static const int BishopPositionW[] =
{3, 0, 0, 0, 0, 0, 0, 3,  
   0, 3, 0, 0, 0, 0, 3, 0,
   0, 0, 3, 0, 0, 3, 0, 0,
   0, 0, 0, 3, 3, 0, 0, 0,
   0, 0, 0, 3, 3, 0, 0, 0,
   0, 0, 3, 0, 0, 3, 0, 0,
   0, 3, 0, 0, 0, 0, 3, 0,
   -3, -6, -6, -6, -6, -6, -6, -3};


static const int PawnCenterScores[] =
{0, 0, 0, 0, 0, 0, 0, 0,
   0, 0, 0, 0, 0, 0, 0, 0,
   0, 0, 0, 0, 0, 0, 0, 0,
   0, 0, 0, 4, 4, 0, 0, 0,
   0, 0, 5, 7, 7, 5, 0, 0,
   0, 0, 3, 4, 4, 3, 0, 0,
   0, 0, -1, -2, -2, -1, 0, 0,
   0, 0, 0, 0, 0, 0, 0, 0};

static const int KingCenterScores[] =
{0, 0, 1, 1, 1, 1, 0, 0,
   0, 1, 2, 3, 3, 2, 1, 0,
   1, 2, 6, 8, 8, 6, 2, 1,
   1, 3, 8, 10, 10, 8, 3, 1,
   1, 3, 8, 10, 10, 8, 3, 1,
   1, 2, 6, 8, 8, 6, 2, 1,
   0, 1, 2, 3, 3, 2, 1, 0,
   0, 0, 1, 1, 1, 1, 0, 0};

static const int KingEndgameScoresW[64] =
{ -32, -19, -19, -19, -19, -19, -19, -32,
  -19, -19,  -6,  -6,  -6,  -6, -19, -19,
  -19,  -6,  13,  19,  19,  13,  -6, -19,
  -19,  -6,  13,  19,  19,  13,  -6, -19,
  -19,  -6,   6,  13,  13,   6,  -6, -19,
  -19,  -6,   0,   6,   6,   0,  -6, -19,
  -19, -13,  -6,   0,   0,  -6, -13, -19,
  -32, -25, -19, -19, -19, -19, -25, -32};

static const int KingEndgameScoresWK[64] =
{ -55, -44, -32, -19, -19, -19, -19, -19,
  -44, -32, -19, -19,  -6,  -6,  -6,  -6,
  -44, -32, -19, -6,    6,  13,  13,   0,
  -44, -32, -19, -6,   13,  19,  19,   0,
  -44, -32, -19, -6,   13,  19,  19,   0,
  -44, -32, -19, -6,    6,  13,  13,   0,
  -44, -32, -19, -13,  -6,  -6,  -6,  -6,
  -55, -32, -32, -19, -19, -13, -13, -19
};

static const int KingEndgameScoresWQ[64] =
{ -19, -19, -19, -19, -19, -32, -44, -55,
   -6,  -6,  -6,  -6, -19, -19, -32, -44, 
    0,  13,  13,   6,  -6, -19, -32, -44,
    0,  19,  19,  13,  -6, -19, -32, -44,
    0,  19,  19,  13,  -6, -19, -32, -44,
    0,  13,  13,   6,  -6, -19, -32, -44,
   -6,  -6,  -6,  -6, -13, -19, -32, -44,
  -19, -13, -13, -19, -19, -32, -32, -55
};

static const int KingEndgameScoresB[64] =
{ -32, -25, -19, -19, -19, -19, -25, -32,
  -19, -13,  -6,   0,   0,  -6, -13, -19,
  -19,  -6,   0,   6,   6,   0,  -6, -19,
  -19,  -6,   6,  13,  13,   6,  -6, -19,
  -19,  -6,  13,  19,  19,  13,  -6, -19,
  -19,  -6,  13,  19,  19,  13,  -6, -19,
  -19, -19,  -6,  -6,  -6,  -6, -19, -19,
  -32, -19, -19, -19, -19, -19, -19, -32 };

static const int KingEndgameScoresBK[64] =
{ 
  -55, -32, -32, -19, -19, -13, -13, -19,
  -44, -32, -19, -13,  -6,  -6,  -6,  -6,
  -44, -32, -19, -6,    6,  13,  13,   0,
  -44, -32, -19, -6,   13,  19,  19,   0,
  -44, -32, -19, -6,   13,  19,  19,   0,
  -44, -32, -19, -6,    6,  13,  13,   0,
  -44, -32, -19, -19,  -6,  -6,  -6,  -6,
  -55, -44, -32, -19, -19, -19, -19, -19,
};

static const int KingEndgameScoresBQ[64] = 
{
  -19, -13, -13, -19, -19, -32, -32, -55,
   -6,  -6,  -6,  -6, -13, -19, -32, -44,
  -19,  -6,  13,  19,  19,  13,  -6, -19,
  -19,  -6,  13,  19,  19,  13,  -6, -19,
  -19,  -6,   6,  13,  13,   6,  -6, -19,
  -19,  -6,   6,  13,  13,   6,  -6, -19,
  -19, -13,  -6,   0,   0,  -6, -13, -19,
  -32, -25, -19, -19, -19, -19, -25, -32
};

static const int CenterPawnCoverInit[] =
{0, 0, 0, 0, 0, 0, 0, 0,
   0, 0, 0, 0, 0, 0, 0, 0,
   0, 0, 0, 0, 0, 0, 0, 0,
   0, 0, 0, 0, 0, 0, 0, 0,
   0, 0, 0, 0, 0, 0, 0, 0,
   0, 0, 1, 1, 1, 1, 0, 0,
   0, 0, 1, 1, 1, 1, 0, 0,
   0, 0, 0, 0, 0, 0, 0, 0};

static const int QueenCenterScores[] =
{-4, 0, 0, 0, 0, 0, 0, -4,
   -2, 0, 0, 0, 0, 0, 0, -2,
   -2, 0, 1, 2, 2, 1, 0, -2,
   -2, 0, 2, 3, 3, 2, 0, -2,
   -2, 0, 2, 3, 3, 2, 0, -2,
   -2, 0, 1, 2, 2, 1, 0, -2,
   -2, 0, 0, 0, 0, 0, 0, -2,
   -4, 0, 0, 0, 0, 0, 0, -4};

static const int OutpostScoresB[] =
{0, 0, 0, 0, 0, 0, 0, 0,
0, 0, 0, 0, 0, 0, 0, 0,
0, 0, 0, 0, 0, 0, 0, 0,
0, 0, 0, 2, 2, 0, 0, 0,
0, 0, 2, 5, 5, 2, 0, 0,
0, 0, 3, 6, 6, 3, 0, 0,
0, 0, 1, 2, 2, 1, 0, 0,
0, 0, 0, 0, 0, 0, 0, 0 };

static const int OutpostScoresW[] =
{0, 0, 0, 0, 0, 0, 0, 0,
0, 0, 1, 2, 2, 1, 0, 0,
0, 0, 3, 6, 6, 3, 0, 0,
0, 0, 2, 5, 5, 2, 0, 0,
0, 0, 0, 2, 2, 0, 0, 0,
0, 0, 0, 0, 0, 0, 0, 0,
0, 0, 0, 0, 0, 0, 0, 0,
0, 0, 0, 0, 0, 0, 0, 0 };

static const int KBNKScores[] =
{-20, -15, -10, -3, 3, 10, 15, 20,
   -15, -10, -3, 0, -1, 3, 10, 15,
   -10, -3, 0, 0, 0, -1, 3, 10,
   -3, -3, 0, -2, -2, 0, -1, 3,
   3, -1, 0, -2, -2, 0, 0, -3,
   10, 3, -1, 0, 0, 0, -3, -10,
   15, 10, 3, -1, 0, -3, -10, -15,
   20, 15, 10, 3, -3, -10, -15, -20};

static const int quadrants[] =
{ 0, 0, 0, 0, 1, 1, 1, 1,
   0, 0, 0, 0, 1, 1, 1, 1,
   0, 0, 0, 0, 1, 1, 1, 1,
   0, 0, 0, 0, 1, 1, 1, 1,
   2, 2, 2, 2, 3, 3, 3, 3,
   2, 2, 2, 2, 3, 3, 3, 3,
   2, 2, 2, 2, 3, 3, 3, 3,
   2, 2, 2, 2, 3, 3, 3, 3};

static int endgame;
static ColorType side_to_move;

// Weights for positional scoring components.  A plus value means that
// the item being scored is valuable to the player; a minus value means
// that a penalty is extracted.  In most cases, scoring components are
// simply summed together to get a total score, but some are folded
// in, in a more complex way.

// material terms
static const int TWO_BISHOPS = 20;
static const int BAD_TRADE=-75;
// development terms
static const int BLOCKED_BISHOP = -4;
static const int BISHOP_MOBILITY[] = {-8, -8, -6, -3, 0, 2, 4, 6, 8, 8, 8 };
static const int BAD_BISHOP = -5;
static const int CENTER_PAWN_BLOCK = -6;  // e or d pawn blocked by piece
static const int LONG_DIAG = 3; // bishop on long diagonal
static const int KNIGHT_BACK = -7;
static const int KNIGHT_OUTPOST = 2; // in addition to reg. outpost scorea
static const int KNIGHT_MOBILITY[] = {-6, -1, 1, 2, 3, 4, 5, 6, 7};
static const int ROOK_ON_7TH_RANK = 15;
static const int LOW_ROOK_MOBILITY = -8;
static const int QUEEN_ON_7TH_RANK = 6;
static const int QUEEN_DEVELOP[] = {6, 6, 6, 5, 5, 5, 4, 4, 4, 3, 3,
   2, 2};
static const int QUEEN_OUT = -7; // premature development of queen
static const int DOUBLED_ROOKS = 8; // on open file
static const int ROOK_ON_OPEN_FILE = 6;
static const int QUEEN_ON_OPEN_FILE = 4;
// castling terms
static const int KCASTLE = 11;
static const int QCASTLE = 8;
static const int CAN_CASTLE = -2;
static const int CANT_CASTLE = -12;
static const int CANT_CASTLE_K_SIDE = -6;
static const int CANT_CASTLE_Q_SIDE = -4;
// pawn structure terms
static const int WEAK = -4;
static const int WEAK2 = -4;
static const int WEAK_ON_OPEN_FILE = -8;
static const int SPACE=2;
// isolated pawn 
//static const int ISOLATED_PAWN[] = { 0, -6, -15, -23, -32, -41, -51, -62 };
static const int ISOLATED_PAWN[] = { 0, -6, -14, -24, -36, -46, -56, -66 };

static const int DOUBLED_PAWNS = -10;
static const int TRIPLED_PAWNS = -5;
static const int ISOLATED_DOUBLED = -10;

static const int PASSED_PAWN[8] = { 0, 10, 20, 30, 45, 65, 95, 0}; // score by rank
//static const int PASSED_PAWN[8] = { 0, 6, 13, 27, 40, 55, 70, 0}; // score by rank
static const int HIDDEN_PASSED_PAWN = 15;
static const int PASSED_PAWN_BLOCK[8] =
//{ 0, 6/3, 13/3, 23/3, 35/3, 51/3, 5, 0 };
{ 0, 10/3, 20/3, 30/3, 45/3, 65/3, 5, 0 };
// outside passed pawns 
//static const int OUTSIDE_PP[] = { 35, 28, 23, 17, 13, 10,
//   7, 5, 4, 3, 3, 3, 3};
static const int OUTSIDE_PP[] = { 35, 35, 35, 32, 27, 24,
  21, 18, 15, 15, 12, 12 };

static const int SUPPORTED_PASSER6 = 24;
static const int SUPPORTED_PASSER7 = 48;
static const int PASSER_MATERIAL_BONUS[] =
 { 12, 8, 5, 3, 2, 1, 0, -2, -4, -6, -6, -6, -6, -6 };
static const int PASSED_PAWN_CANT_ADVANCE = 4;
static const int ROOK_BEHIND_PP = 25;
static const int ROOK_AHEAD_OF_PP = -10;
static const int CONNECTED_PASSERS[8] = 
//{0, 6, 7, 8, 10, 16, 24, 0};
{0, 6, 7, 8, 20, 37, 75, 0};
static const int ADJACENT_PASSERS = 8;
static const int ADV_CONNECTED_PASSERS = 25;
static const int RIM_PASSED_PAWN = -7; // pawn on h-file or a-file
static const int PAWN_DUO = 5; // a la Kmoch
static const int ADV_UNSUPPORTED_PAWN = -4; // an advanced but unsupported pawn
static const int STONEWALL = -8; // opponent has stonewall formation

// king safety terms
static const int KING_OFF_BACK_RANK = -20;
static const int KS_MATERIAL_SCALE[] =
  { 0, 0, 1, 2, 4, 7, 9, 11, 13, 15, 16, 16, 16 };
static const int FIANCHETTO = 6;
static const int BP_COVER_MISSING = -8;
static const int NP_COVER_MISSING = -9;
static const int RP_COVER_MISSING = -9;
static const int BP_COVER_MISSING2 = -15;
static const int NP_COVER_MISSING2 = -16;
static const int RP_COVER_MISSING2 = -16;
static const int BP_COVER_MISSING3 = -19;
static const int NP_COVER_MISSING3 = -20;
static const int RP_COVER_MISSING3 = -20;
static const int BP_COVER_MISSING4 = -24;
static const int NP_COVER_MISSING4 = -25;
static const int RP_COVER_MISSING4 = -25;
static const int KING_NEAR_OPEN_FILE = -8;
static const int DOUBLED_PAWN_NEAR_KING = 5;
static const int MIDDLE_OPEN = -4;
static const int KS_DEFECTS[] = {-2, -5, -9, -14, -20, -20, -20, -20, -20, -20, -20 };

static const int PAWN_WALL = -2;
static const int KING_QUEEN_DISTANCE[] = {3, 3, 2, 1, 0, 0, 0, 0};

static const int PROXIMITY_SCALE[] = 
  { 0, 5, 12, 19, 26, 36, 46, 56, 66, 77, 88, 99, 107,
   107, 107, 107, 107, 107, 107, 107, 107, 107, 107, 107, 107 };

static const int KING_TRAPPED_IN_CORNER = -5; 
// king position terms - valid only in endgame
static const int ADVANCED_KING = 4;
static const int OPPOSITION = 6; // twice this for direct opposition
// piece safety terms
static const int BISHOP_TRAPPED = -100; // bishop trapped by pawn
// general endgame terms
static const int KING_PAWN_DISTANCE = -10; // distance from opposing king
// KP vs. K endgame
static const int UNCATCHABLE_PAWN = 250;
// a pawn that may become "uncatchable"
static const int UNCATCHABLE_PAWN2[] = {24, 24, 18, 12, 8, 4, 0, 0, 0, 0,
0, 0, 0};
static const int KING_AHEAD_OF_PAWN = 16;
static const int DISTANCE_FROM_PAWN = -4;
// Bishop vs. Knight with pawns on both sides
static const int BISHOP_ENDGAME_BONUS[] = { 21, 21, 21, 21, 15, 10 };

static Bitmap passedW[64];
static Bitmap passedB[64];
static Bitmap backwardW[64];
static Bitmap backwardB[64];
static Bitmap weakW[64];
static Bitmap weakB[64];
static Bitmap outpostW[64];
static Bitmap outpostB[64];
static Bitmap right_side_mask[8], left_side_mask[8];
// area near king 
static Bitmap king_attacks[4];
static Bitmap king_proximity[64];
static Bitmap king_file_proximity[8];
static Bitmap king_knight_proximity[64];
static Bitmap w_pawn_attacks[8], b_pawn_attacks[8];
static Bitmap center_squares;
static Bitmap white_squares;
static Bitmap black_squares;
static Bitmap center_pawn_cover_b;
static Bitmap center_pawn_cover_w;
static Bitmap rook_pawn_mask;
static Bitmap longdiagDa1,longdiagDa8;
static Bitmap w_bishop_trap, b_bishop_trap;
static Bitmap w_stonewall, b_stonewall;
static Bitmap right_mask,left_mask, abcd_mask, efgh_mask;
static Bitmap w_king_pos1[64], w_king_pos2[64], b_king_pos1[64], b_king_pos2[64];
static Bitmap w_files_ahead[64],b_files_ahead[64];
static int done_init = 0;
static Pawn_Entry *pawn_hash_table = NULL;
static King_Pawn_Entry *king_pawn_hash_table = NULL;
static Bitmap ghfiles;
static int opposition[64][64];
int distances[64][64];

static int b_material_level, w_material_level;
static Pawn_Entry *pawn_hash_entry;
static Pawn_Entry::PawnData *hash_entry_w, *hash_entry_b;
// most advanced passer that is "uncatchable". 
static int white_passer_rank, black_passer_rank;
// most advanced passer that is "uncatchable" (do not count
// separated pawns that are not individually uncatchable)
static int white_passer_race_rank, black_passer_race_rank;
static Square white_passer_sq, black_passer_sq;
static Square white_passer_race_sq, black_passer_race_sq;

// Don't call it "distance" - conflicts with MS template library
static inline int distance1(int x, int y)
{
   return distances[x][y];
}

// Return all squares attacked by a pawn of color "side" at square "sq".
#define pawn_attacks( board, sq, side) \
   Bitmap::And(Bearing::pawn_attacks[sq][side],board.pawn_bits[side])

#define PAWN_HASH_SIZE 15000
#define KING_PAWN_HASH_SIZE 15000

static void init_bitmaps()
{
   int i, r,f;

   for (i = 0; i < 64; i++)
   {
      for (int j = 0; j < 64; j++)
      {
         int file_distance, rank_distance;
         Square kp = i;
         Square oppkp = j;
         file_distance=Util::Abs(File(kp)-File(oppkp));
         rank_distance=Util::Abs(Rank(kp,Black)-Rank(oppkp,Black));
         if (rank_distance == 0 && file_distance == 2)
         {
            opposition[i][j] = 2; // direct opposition
         } else if (file_distance == 0 && rank_distance % 2 == 0)
            opposition[i][j] = 1; // indirect opposition
         else if ((file_distance == 2) && (rank_distance == 2))
            opposition[i][j] = 1;
         else
            opposition[i][j] = 0;
         if (file_distance > rank_distance)
            distances[i][j] = file_distance;
         else
            distances[i][j] = rank_distance;
      }
   }
   Square sq;
   for (sq = 0; sq < 64; sq++)
   {
      if (InCenter(sq))
      {
         center_squares.set(sq);
      }
      if (SquareColor(sq) == White)
         white_squares.set(sq);
      else
         black_squares.set(sq);
      if (CenterPawnCoverInit[sq])
      {
         center_pawn_cover_w.set(sq);
         center_pawn_cover_b.set(63-sq);
      }
      int rank = Rank(sq,White);
      int rankb= Rank(sq,Black);

      int file = File(sq);
      if (file == 1 || file == 8)
         rook_pawn_mask.set(sq);
      for (r = rank; r > 1; r--)
      {
         Square sq2 = Square(file,r,White);
         if (file != 8)
         {
            sq2 = Square(file+1,r,White);
            backwardW[sq].set((int)sq2);
         }
         if (file != 1)
         {
            sq2 = Square(file-1,r,White);
            backwardW[sq].set((int)sq2);
         }
      }
      int x = 0;
	  if (rank<8) {
      for (r=rank+1;r<=8;r++) {
         Square sq2 = Square(file,r,White);
         w_files_ahead[sq].set(sq2);
         Square sq3 = sq2;
         while (file-File(sq3)<=x) {
           b_king_pos1[sq].set(sq3);
		   if (File(sq3)==1) break;
		   else sq3--;
         }
         if (x==0 && File(sq) != 1) b_king_pos1[sq].set(sq2-1);
         sq3 = sq2;
         while (File(sq3)-file<=x) {
           b_king_pos1[sq].set(sq3); 
		   if (File(sq3)==8) break;
		   else sq3++;
         }
         if (x==0 && File(sq) != 8) b_king_pos1[sq].set(sq2+1);
		 int y = x-1;
		 sq3 = sq;
         while (file-File(sq3)<=y) {
           b_king_pos2[sq].set(sq3); 
		   if (File(sq3)==1) break;
		   else sq3--;
         }
         sq3 = sq2;
         while (File(sq3)-file<=7) {
           b_king_pos2[sq].set(sq3); 
		   if (File(sq3)==8) break;
		   else sq3++;
         }
         x++;
      }
	  }
      x = 0;         
	  if (rankb<8) {
      for (r=rankb+1;r<=8;r++) {
         Square sq2 = Square(file,r,Black);
         b_files_ahead[sq].set(sq2);
         Square sq3 = sq2;
 	 w_king_pos1[sq].set(sq2);
         while (file-File(sq3)<=x) {
           w_king_pos1[sq].set(sq3); 
		   if (File(sq3)==1) break;
		   else sq3--;
         }
         if (x==0 && File(sq) != 1) w_king_pos1[sq].set(sq2-1);
         sq3 = sq2;
         while (File(sq3)-file<=x) {
           w_king_pos1[sq].set(sq3); 
		   if (File(sq3)==8) break;
		   else sq3++;
         }
         if (x==0 && File(sq) != 8) w_king_pos1[sq].set(sq2+1);
	 int y = x-1;
         while (file-File(sq3)<=y) {
           w_king_pos2[sq].set(sq3); 
		   if (File(sq3)==1) break;
		   else sq3--;
         }
         sq3 = sq2;
         while (File(sq3)-file<=y) {
           w_king_pos2[sq].set(sq3); 
		   if (File(sq3)==8) break;
		   else sq3++;
         }
         x++;
      }
  }

      weakW[sq] = backwardW[sq];
      if (file != 8)
      {
         weakW[sq].clear(Square(file+1,rank,White));
      }
      if (file != 1)
         weakW[sq].clear(Square(file-1,rank,White));
      for (r = rank + 1; r < 8; r++)
      {
         Square sq2 = Square(file,r,White);
         passedW[sq].set(sq2);
         if (file != 8)
         {
            sq2 = Square(file+1,r,White);
            passedW[sq].set(sq2);
            outpostW[sq].set(sq2);
         }
         if (file != 1)
         {
            sq2 = Square(file-1,r,White);
            passedW[sq].set(sq2);
            outpostW[sq].set(sq2);
         }
      }
      rank = Rank(sq,Black);
      for (r = rank; r > 1; r--)
      {
         Square sq2 = Square(file,r,Black);
         if (file != 8)
         {
            sq2 = Square(file+1,r,Black);
            backwardB[sq].set(sq2);
         }
         if (file != 1)
         {
            sq2 = Square(file-1,r,Black);
            backwardB[sq].set(sq2);
         }
      }
      weakB[sq] = backwardB[sq];
      if (file != 8)
         weakB[sq].clear(Square(file+1,rank,Black));
      if (file != 1)
         weakB[sq].clear(Square(file-1,rank,Black));
      for (r = rank + 1; r < 8; r++)
      {
         Square sq2 = Square(file,r,Black);
         passedB[sq].set((int)sq2);
         if (file != 8)
         {
            sq2 = Square(file+1,r,Black);
            passedB[sq].set(sq2);
            outpostB[sq].set(sq2);
         }
         if (file != 1)
         {
            sq2 = Square(file-1,r,Black);
            passedB[sq].set(sq2);
            outpostB[sq].set(sq2);
         }
      }
   }
   for (r = 1; r <= 4; r++)
   {
      for (f = 1; f<=4; f++)
      {
         Square sq = Square(f,r,Black);
         king_attacks[0].set(sq);
      }
   }
   for (r = 1; r <= 4; r++)
   {
      for (f = 5; f<=8; f++)
      {
         Square sq = Square(f,r,Black);
         king_attacks[1].set(sq);
      }
   }
   for (r = 1; r <= 4; r++)
   {
      for (f = 1; f<=4; f++)
      {
         Square sq = Square(f,r,White);
         king_attacks[2].set(sq);
      }
   }
   for (r = 1; r <= 4; r++)
   {
      for (f = 5; f<=8; f++)
      {
         Square sq = Square(f,r,White);
         king_attacks[3].set(sq);
      }
   }
   for (sq = 0; sq < 64; sq++) {
      king_proximity[sq] = king_attacks[quadrants[sq]];
      Bitmap dests(Bearing::king_attacks[sq]);
      dests.set(sq);
      Square sq2;
      while ((sq2 = dests.first_one()) != InvalidSquare) {
        king_knight_proximity[sq].Or(
           Bearing::knight_attacks[sq2].data);
        dests.clear(sq2);
      }
   }

   for (int file = 1; file < 9; file++)
   {
      for (int r = 2; r < 5; r++)
      {
          w_pawn_attacks[file-1].set((int)Square(Util::Max(1,file-1),r,White));
          w_pawn_attacks[file-1].set((int)Square(file,r,White));
          w_pawn_attacks[file-1].set((int)Square(Util::Min(8,file+1),r,White));
          if (file==8) w_pawn_attacks[file-1].set((int)Square(6,r,White));
          if (file==6) w_pawn_attacks[file-1].set((int)Square(8,r,White));
          if (file==1) w_pawn_attacks[file-1].set((int)Square(3,r,White));
          if (file==3) w_pawn_attacks[file-1].set((int)Square(1,r,White));
          b_pawn_attacks[file-1].set((int)Square(Util::Max(1,file-1),r,Black));
          b_pawn_attacks[file-1].set((int)Square(file,r,Black));
          b_pawn_attacks[file-1].set((int)Square(Util::Min(8,file+1),r,Black));
          if (file==8) b_pawn_attacks[file-1].set((int)Square(6,r,Black));
          if (file==6) b_pawn_attacks[file-1].set((int)Square(8,r,Black));
          if (file==1) b_pawn_attacks[file-1].set((int)Square(3,r,Black));
          if (file==3) b_pawn_attacks[file-1].set((int)Square(1,r,Black));
      }
   }
   longdiagDa1= Bitmap(0);
   int x;
   for (x = 56; x >= 0; x-=7)
      longdiagDa1.set(x);
   longdiagDa8= Bitmap(0);
   for (x = 0; x <= 63; x+=9)
      longdiagDa8.set(x);
   for (x=1; x <8; x++)
   {
      for (int y=x-1;y>=0;y--)
         left_side_mask[x].Or(Bearing::file_mask[y]);
   }
   for (x=6; x >=0; x--)
   {
      for (int y=x+1;y<8;y++)
         right_side_mask[x].Or(Bearing::file_mask[y]);
   }
   for (x = 0; x < 8; x++) {
      king_file_proximity[x] = Bearing::file_mask[x];
      if (x == 0)
        king_file_proximity[x].data |= Bearing::file_mask[2].data;
      else
        king_file_proximity[x].data |= Bearing::file_mask[x-1].data;
      if (x==7)
        king_file_proximity[x].data |= Bearing::file_mask[5].data;
      else              
        king_file_proximity[x].data |= Bearing::file_mask[x+1].data;
   }
   b_bishop_trap.set(H2);
   b_bishop_trap.set(G1);
   b_bishop_trap.set(A2);
   b_bishop_trap.set(B1);
   w_bishop_trap.set(B8);
   w_bishop_trap.set(A7);
   w_bishop_trap.set(G8);
   w_bishop_trap.set(H7);
   w_stonewall.set(D4);
   w_stonewall.set(F4);
   b_stonewall.set(D5);
   b_stonewall.set(F5);
   ghfiles = Bitmap::Or(Bearing::file_mask[7],Bearing::file_mask[6]);
   right_mask = ghfiles;
   right_mask.Or(Bearing::file_mask[5]);
   left_mask = Bitmap::Or(Bearing::file_mask[0],Bearing::file_mask[1]);
   left_mask.Or(Bearing::file_mask[2]);
   abcd_mask = left_mask;
   abcd_mask.Or(Bearing::file_mask[3]);
   efgh_mask = right_mask;
   efgh_mask.Or(Bearing::file_mask[4]);
}

static FORCEINLINE int FileOpen( const int file, const ColorType side ) 
{
   return ((side == White) ? hash_entry_w->open_files :
                            hash_entry_b->open_files) & (1<<(file-1));
}

static FORCEINLINE int FileOpen( const int file )
{
    return (hash_entry_w->open_files & hash_entry_b->open_files) & (1<<(file-1));
}

int uncatchable(const Board &board,
                 const Square pawnpos,
                 const Square kp,
                 const Square oppkp,
                 const ColorType side)
{
   // Use the "square of the king" rule to see whether a passed pawn can
   // be captured before it promotes (or immediately after).

   int oppkp_rank = Rank(oppkp,side);
   int rankdist = Rank(pawnpos,side) - oppkp_rank;
   int sqkingdist;
   if (board.Side() == side)
   {
      if (rankdist > 0)
      {
         // If the king is behind us and we are on move we can't be caught.
         return 1 /*TRUE*/;
      } 
      else
      {
         sqkingdist = distance1(oppkp,Square(File(pawnpos),8,side))-(8-Rank(pawnpos,side));
         if (sqkingdist>0)
            return 1 /*TRUE*/;
      }
   } 
   else
   {
      if (rankdist > 1)
      {
         // pawn > 1 square ahead
         return 1 /*TRUE*/;
      } 
      else
      {
         sqkingdist = distance1(oppkp,Square(File(pawnpos),8,side))-(9-Rank(pawnpos,side));
         if (sqkingdist>0)
            return 1 /*TRUE*/;
      }
   }
   // The opposing king could theoretically catch the pawn.  Now
   // see if our king is in a position to prevent this.
   int kpfile = File(kp);
   int pawnfile = File(pawnpos);
   int kprank = Rank(kp,side);
   int pawnrank = Rank(pawnpos,side);
   if (kprank < pawnrank)
   {
      // our king is behind our pawn
      return 0 /*FALSE*/;
   }
   int d;
   if (pawnfile == 1)
   {
      // rook pawn.
      if (kpfile == 1 && File(oppkp) == 3)
      {
         return (board.Side() == side && pawnrank < 7 &&
             kprank > oppkp_rank &&
            (kprank-pawnrank <= 3));
      }
      if (File(oppkp) == 1 || 
          distance1(oppkp,pawnpos) - distance1(kp,pawnpos) < 0)
      {
         // opposing king is ahead of the pawn, or closer
         // than our own king.
         return 0 /*FALSE*/;
      }
      if (side == White)
      {
         if (oppkp <= C8)
            return 0 /*FALSE*/;
         d = distance1(oppkp,C8) - distance1(kp,B7);
         if (board.Side() == side && d >= 0)
         {
             return 1 /*TRUE*/;
         }
         if (board.Side() != side && d > 0)
         {
            return 1 /*TRUE*/;
         }
      } 
      else
      {
         if (oppkp >= A1 && oppkp <= C1)
            return 0 /*FALSE*/;
         d = distance1(oppkp,C1) - distance1(kp,B2);
         if (board.Side() == side && d >= 0)
         {
             return 1 /*TRUE*/;
         }
         if (board.Side() != side && d > 0)
         {
            return 1 /*TRUE*/;
         }
      }
   } 
   else if (pawnfile == 8)
   {
      if (kpfile == 8 && File(oppkp) == 6)
      {
         return (board.Side() == side && pawnrank < 7 &&
             kprank > oppkp_rank &&
            (kprank-pawnrank <= 3));
      }
      if (File(oppkp) == 8 || 
          distance1(oppkp,pawnpos) - distance1(kp,pawnpos) < 0)
      {
         // opposing king is ahead of the pawn, or closer
         // than our own king.
         return 0 /*FALSE*/;
      }
      if (side == White)
      {
         if (oppkp >= F8 && oppkp <= H8)
            return 0 /*FALSE*/;
         d = distance1(oppkp,F8) - distance1(kp,G7);
         if (board.Side() == side && d >= 0)
         {
             return 1 /*TRUE*/;
         }
         if (board.Side() != side && d > 0)
         {
            return 1 /*TRUE*/;
         }
      } 
      else
      {
         if (oppkp >= F1)
            return 0 /*FALSE*/;
         d = distance1(oppkp,F1) - distance1(kp,G2);
         if (board.Side() == side && d >= 0)
         {
             return 1 /*TRUE*/;
         }
         if (board.Side() != side && d > 0)
         {
            return 1 /*TRUE*/;
         }
      }
   } 
   else
   {
          // not a rook pawn
         // If our king is >=2 squares ahead and the opposing king
         // is not closer, it is a win
         if (distance1(kp,pawnpos) <= distance1(oppkp,pawnpos))
         {
            if (kprank > pawnrank + 1)
               return 1 /*TRUE*/;
         }
         // a supported pawn on the 6th rank or past is uncatchable
         if (pawnrank >=6  && kprank >= 6)
         {
            Bitmap ka(Bearing::king_attacks[pawnpos]);
            if (ka.is_set(board.KingPos(side)))
               return 1 /*TRUE*/;
         }
		 int guarded = (kpfile == pawnfile-1 ||
              kpfile == pawnfile+1) && (kprank == pawnrank ||
              kprank == pawnrank+1);
         // If the king is directly ahead of the pawn, or on an adjacent file,
		 // and we have the opposition, it is a win
         if (board.Side() != side && (guarded || kpfile==pawnfile) && kprank == pawnrank+1)
         {
            if (File(oppkp) == pawnfile &&
                ((int)oppkp - (int)kp) % 16 == 0)
               return 1 /*TRUE*/;
         }
         // If we guard the pawn and the opposing king is behind or
         // on the same rank, or far enough away, we win
         if (guarded) {
			if (rankdist>=0) return 1;
			else if (kprank == pawnrank+1)
			  return 1;
			else if (board.Side() == side) {
              return ((side==White) ?
               !b_king_pos2[pawnpos].is_set(oppkp) : !w_king_pos2[pawnpos].is_set(oppkp));
			}
			else
              return ((side==White) ?
               !b_king_pos1[pawnpos].is_set(oppkp) : !w_king_pos1[pawnpos].is_set(oppkp));
         }
   }
   return 0 /*FALSE*/;
}   

// evaluate outside passed pawns.
static void eval_outside_pp(const Board &board,Pawn_Entry *pawn_hash_entry)
{
   // This code is based on similar code in Crafty.  See also
   // Fine's Basic Chess Endings, pp. 35-36.

   byte first_passer_w = pawn_hash_entry->w_pawn_entry.first_passer;
   byte last_passer_w = pawn_hash_entry->w_pawn_entry.last_passer;
   byte first_passer_b = pawn_hash_entry->b_pawn_entry.first_passer;
   byte last_passer_b = pawn_hash_entry->b_pawn_entry.last_passer;

   pawn_hash_entry->b_pawn_entry.outside = 
      pawn_hash_entry->w_pawn_entry.outside = 0;
   if (first_passer_w == 9 && first_passer_b == 9) {
      return;
   }
   Bitmap all_pawns(Bitmap::Or(board.pawn_bits[White],board.pawn_bits[Black]));
   if (first_passer_w != 9)
   {
      if ((first_passer_w < first_passer_b-1) ||
          (last_passer_w > last_passer_b+1))
      {
         if (first_passer_w <= 4)
         {
            if (!Bitmap::And(all_pawns,right_side_mask[first_passer_w-1]).is_clear()
                &&  Bitmap::And(board.pawn_bits[Black],left_side_mask[first_passer_w-1]).is_clear())
            {
               if (!pawn_hash_entry->w_pawn_entry.protected_pp || pawn_hash_entry->b_pawn_entry.protected_pp)
               {
                  pawn_hash_entry->w_pawn_entry.outside++;
                  goto step2;
               }
            }
         }
         if (last_passer_w >= 5)
         {
            if (!Bitmap::And(all_pawns,left_side_mask[last_passer_w-1]).is_clear()
                &&  Bitmap::And(board.pawn_bits[Black],right_side_mask[last_passer_w-1]).is_clear())
            {
               if (!pawn_hash_entry->w_pawn_entry.protected_pp || pawn_hash_entry->b_pawn_entry.protected_pp)
               {
                  pawn_hash_entry->w_pawn_entry.outside++;
               }
            }
         }
      }
   }
   step2:
   if (first_passer_b != 9)
   {
      if ((first_passer_b < first_passer_w-1) ||
          (last_passer_b > last_passer_w+1))
      {
         if (first_passer_b <= 4)
         {
            if (!Bitmap::And(all_pawns,right_side_mask[first_passer_b-1]).is_clear()
                &&  Bitmap::And(board.pawn_bits[White],left_side_mask[first_passer_b-1]).is_clear())
            {
               if (!pawn_hash_entry->w_pawn_entry.protected_pp || pawn_hash_entry->w_pawn_entry.protected_pp)
               {
                  pawn_hash_entry->b_pawn_entry.outside++;
                  goto step3;
               }
            }
         }
         if (last_passer_b >= 5)
         {
            if (!Bitmap::And(all_pawns,left_side_mask[last_passer_b-1]).is_clear()
                &&  Bitmap::And(board.pawn_bits[White],right_side_mask[last_passer_b-1]).is_clear())
            {
               if (!pawn_hash_entry->w_pawn_entry.protected_pp || pawn_hash_entry->b_pawn_entry.protected_pp)
               {
                  pawn_hash_entry->b_pawn_entry.outside++;
               }
            }
         }
      }
   }
   step3:
   return;
}

void Scoring::init(const Board &board)
{
   if (!done_init)
   {
      init_bitmaps();
      done_init++;
   }
   if (!pawn_hash_table)
   {
      // fixed size for now
      pawn_hash_table = (Pawn_Entry*)(new char[PAWN_HASH_SIZE*sizeof(Pawn_Entry)]);
      memset(pawn_hash_table,0xff,PAWN_HASH_SIZE*sizeof(Pawn_Entry));
      king_pawn_hash_table = (King_Pawn_Entry*)(new char[KING_PAWN_HASH_SIZE*sizeof(King_Pawn_Entry)]);
      memset(king_pawn_hash_table,0xff,KING_PAWN_HASH_SIZE*sizeof(King_Pawn_Entry));
  }
  side_to_move = board.Side();
}      


void Scoring::cleanup()
{
   if (done_init)
   {
      delete [] pawn_hash_table;
      delete [] king_pawn_hash_table;
   }
}

int Scoring::in_endgame( const Board &board )
{
     const Material & bmat = board.getMaterial(Black);
     const Material & wmat = board.getMaterial(White);
     return bmat.material_level() <= 7 &&
            wmat.material_level() <= 7 &&
            (bmat.piece_count() - bmat.pawn_count()/2 <= 7 &&
             wmat.piece_count() - wmat.pawn_count()/2 <= 7);
}

// adjust the material score. "ourmat" is the side with the
// material advantage
static int adjust_mat_score(int mdiff, const Material &ourmat,
                            const Material &oppmat)
{  
   int nonpawns = 
   (int)(ourmat.value() - ourmat.pawn_count()*64);
   if (nonpawns < 4000)
   {
      int max = 80 - ((nonpawns-2000)/25);
      return mdiff + (max*mdiff)/(9*64);
   }
   return mdiff;
}

static int KBPDrawW(const Board &board) {
   Bitmap pb(board.pawn_bits[White]);
   if (!(Bitmap::And(pb,Bearing::file_mask[0])).is_clear())
   {
       // on the a file
       // queen square is white
       if ((Bitmap::And(board.bishop_bits[White],Bearing::white_squares)).is_clear())
       {
          // wrong color bishop
          Square pawnsq = pb.first_one();
          if (Rank(board.KingPos(Black),White) >= Rank(pawnsq,White))
          {
             if (Util::Abs((int)File(board.KingPos(Black)) - (int)File(pawnsq)) <= 1)
             {
                return 1 /*TRUE*/;
             }
             Square queenSq = Square(AFILE,8,White);
             if ((distance1(queenSq,board.KingPos(Black)) < distance1(pawnsq,queenSq)) &&
                 distance1(queenSq,board.KingPos(Black)) < distance1(queenSq,board.KingPos(White)))
             {
                return 1;
             }
          }
       }
    } 
    else if (!(Bitmap::And(pb,Bearing::file_mask[7])).is_clear())
    {
       // on the h file
       // queen square is black
       if ((Bitmap::And(board.bishop_bits[White],Bearing::black_squares)).is_clear())
       {
           // wrong color bishop
           Square pawnsq = pb.first_one();
           if (Rank(board.KingPos(Black),White) >= Rank(pawnsq,White))
           {
              if (Util::Abs((int)File(board.KingPos(Black)) - (int)File(pawnsq)) <= 1)
              {
                 return 1 /*TRUE*/;
              }
              Square queenSq = Square(HFILE,8,White);
              if ((distance1(queenSq,board.KingPos(Black)) < distance1(pawnsq,queenSq)) &&
                  distance1(queenSq,board.KingPos(Black)) < distance1(queenSq,board.KingPos(White)))
              {
                 return 1;
              }
           }
       }
    }
    return 0;
}

static int KBPDrawB(const Board &board) {
   Bitmap pb(board.pawn_bits[Black]);
   if (!(Bitmap::And(pb,Bearing::file_mask[0])).is_clear())
   {
      // on the a file
      // queen square is black
      if ((Bitmap::And(board.bishop_bits[Black],Bearing::black_squares)).is_clear())
      {
         // wrong color bishop
         Square pawnsq = pb.first_one();
         if (Rank(board.KingPos(White),Black) >= Rank(pawnsq,Black))
         {
            if (Util::Abs((int)File(board.KingPos(White)) - (int)File(pawnsq)) <= 1)
            {
               return 1 /*TRUE*/;
            }
            Square queenSq = Square(AFILE,8,Black);
            if ((distance1(queenSq,board.KingPos(White)) < distance1(pawnsq,queenSq)) &&
                distance1(queenSq,board.KingPos(White)) < distance1(queenSq,board.KingPos(Black)))
            {
               return 1;
            }
         }
      }
   } 
   else if (!(Bitmap::And(pb,Bearing::file_mask[7])).is_clear())
   {
      // on the h file
      // queen square is white
      if ((Bitmap::And(board.bishop_bits[Black],Bearing::white_squares)).is_clear())
      {
         // wrong color bishop
         Square pawnsq = pb.first_one();
         if (Rank(board.KingPos(White),Black) >= Rank(pawnsq,Black))
         {
            if (Util::Abs((int)File(board.KingPos(White)) - (int)File(pawnsq)) <= 1)
            {
               return 1 /*TRUE*/;
            }
            Square queenSq = Square(HFILE,8,Black);
            if ((distance1(queenSq,board.KingPos(White)) < distance1(pawnsq,queenSq)) &&
                distance1(queenSq,board.KingPos(White)) < distance1(queenSq,board.KingPos(Black)))
            {
               return 1;
            }
         }
      }
   }
   return 0;
}

static int MaterialScore( const Board &board, const ColorType side )
{
   // Computes material score, from the perspective of 'side'.

   const Material &ourmat = board.getMaterial(side);
   const Material &oppmat = board.getMaterial(OppositeColor(side));
   int mdiff =  (int)(ourmat.value() - oppmat.value());
   //
   // Special case code for pawnless endings. Some are sure wins,
   // some are draws (at best) because of insufficient material to mate.
   //
   if (ourmat.no_pawns() && oppmat.no_pawns())
   {
      if (oppmat.king_only() && mdiff > 0)
      {
         if (ourmat.value() >= 2320 &&
             ((unsigned)ourmat.infobits() != Material::KNN)) // not KNN vs K
         {
            return mdiff + mdiff/10; // bonus because we are sure to mate
         } else
            return (mdiff*2)/10; // not enough material to mate
      }
      else if (ourmat.king_only() && mdiff < 0)
      {
         if (oppmat.value() >= 2320 &&
             ((unsigned)oppmat.infobits() != Material::KNN)) // not KNN vs K
         {
            return mdiff - mdiff/10; // bonus because we are sure to mate
         } else
            return (mdiff*2)/10; // not enough material to mate
      }
      else if (!ourmat.no_pieces() && !oppmat.no_pieces()) {
        if (oppmat.value() <= KING_VALUE + BISHOP_VALUE + ROOK_VALUE &&
            ourmat.value() <= KING_VALUE + BISHOP_VALUE + ROOK_VALUE) {
           // Pawnless endings are hard to win, unless we are way
           // ahead.
           return (mdiff*4)/10;
        }
        else if (oppmat.value() <= KING_VALUE + 2*BISHOP_VALUE + ROOK_VALUE &&
            ourmat.value() <= KING_VALUE + 2*BISHOP_VALUE + ROOK_VALUE) {
           return (mdiff*6)/10;
        }
      }
   }
   // If we are ahead in material give a bonus for trading down
   // pieces (but not pawns).
   // But don't do so if we are so low in material that we can't 
   // mate.
   //
   if (ourmat.no_pawns() && mdiff > 0 && (ourmat.value()< KING_VALUE+BISHOP_VALUE ||
        ((unsigned)ourmat.infobits() == Material::KNN)))
   {
       return (mdiff*4)/10; // not enough material to mate
   }
   if (oppmat.no_pawns() && mdiff < 0 && (oppmat.value()< KING_VALUE+BISHOP_VALUE ||
        ((unsigned)oppmat.infobits() == Material::KNN)))
   {
       return (mdiff*4)/10; // not enough material to mate
   }
   // check for "wrong" rook pawn + Bishop. If we are ahead in material
   // but have this configuration, we cannot do better than draw.
   if (ourmat.infobits() == Material::KBP && mdiff > 0) {
      if ((side == White && KBPDrawW(board)) ||
         (side == Black && KBPDrawB(board))) return 0;
   }
   else if (oppmat.infobits() == Material::KBP && mdiff < 0) {
      if ((side == Black && KBPDrawW(board)) ||
         (side == White && KBPDrawB(board))) return 0;
   }
   // Penalize bad trades. Algorithm similar to Crafty's.
   int score = 0;
   int bad_trade = 0;
   int oppMinors = oppmat.bishop_count() + oppmat.knight_count();
   int ourMinors = ourmat.bishop_count() + ourmat.knight_count();
   if (ourMinors != oppMinors)
   {
     int ourQ = ourmat.queen_count();
     int oppQ = oppmat.queen_count();
     if (ourQ == oppQ && ourmat.rook_count() == oppmat.rook_count())
     {
       // penalize knight or bishop traded away for pawns
       if (ourMinors+1 == oppMinors) bad_trade+=BAD_TRADE;
       else if (oppMinors+1 == ourMinors) bad_trade-=BAD_TRADE;
     }
     else if (ourQ != oppQ)
     {  
        // penalize Q traded for Rook + Knight or bishop
        if (ourQ == 0 && ourMinors<=oppMinors+1) 
           bad_trade += BAD_TRADE;
        else if (oppQ == 0 && oppMinors<=ourMinors+1)
           bad_trade -= BAD_TRADE;
     }
     else
     {
       // penalize rook traded for knight+bishop
       if (ourMinors > oppMinors+1) bad_trade-=BAD_TRADE;
       else if (oppMinors > ourMinors+1) bad_trade+=BAD_TRADE;
     }
   }
   if (bad_trade) {
       int x = Util::Min(ourmat.majorPieceCount(),
                        oppmat.majorPieceCount());
       if (x < 3) bad_trade = bad_trade/(4-x);
       // If it looks like we have made a bad trade but we have
       // lots of compensation (e.g. pawns) give reduced penalty
       if (bad_trade < 0 && mdiff >= 3*PAWN_VALUE)
          bad_trade /= 2;
       else if (bad_trade >0 && mdiff <= -3*PAWN_VALUE)
          bad_trade /= 2;
       score += bad_trade;
   }
#ifdef EVAL_DEBUG
   if (bad_trade) cout << "bad trade score = " << bad_trade << endl;
#endif
   // Give a proportional bonus if we are substantially ahead (or
   // a penalty if we are behind) in material (but not if there
   // are zero pawns).
   if (!ourmat.no_pawns() && ourmat.value() > KING_VALUE + BISHOP_VALUE + PAWN_VALUE &&
       mdiff > 2*(int)PAWN_VALUE)
   {
      return score+adjust_mat_score(mdiff,ourmat,oppmat);
   } 
   else if (!oppmat.no_pawns() && oppmat.value() > KING_VALUE + BISHOP_VALUE + PAWN_VALUE &&
       mdiff < -2*(int)PAWN_VALUE)
   {
      return score-adjust_mat_score(-mdiff,oppmat,ourmat);
   } 
   else
      return score+mdiff;
}

static int pawn_structure_score(const Board &board, ColorType side, 
                                Pawn_Entry::PawnData &entr)
{
   // This function computes a pawn structure score that depends
   // only on the location of pawns (of both colors). It also fills
   // in the pawn hash entry.
   //
   entr.score = 0;
   entr.passers_count = 0;
   entr.protected_pp = 0;
   entr.first_passer = 9;
   entr.last_passer = 0;
   entr.center_score = 0;
   entr.weak_files = 0;
   entr.open_files = 0xff;
   entr.w_square_pawns = entr.b_square_pawns = 0;
   entr.pawns_on_2nd_rank = 0;

   int isolated_count = 0;

   int score = 0;
   int incr = (side == White) ? -8 : 8;
   const ColorType oside = OppositeColor(side);
   entr.opponent_pawn_attacks = board.all_pawn_attacks(oside);
   Piece enemy_pawn = MakePiece(Pawn,oside);
   Piece our_pawn = MakePiece(Pawn,side);
   Bitmap bi(board.pawn_bits[side]);
   Bitmap all_pawns(board.pawn_bits[White]);
   all_pawns.Or(board.pawn_bits[Black]);
   Square sq;
   int lside = 0, rside = 0;
#ifdef PAWN_DEBUG
   int total_space = 0;
#endif
   for (;(sq = bi.first_one()) != InvalidSquare;bi.clear(sq))
   {
#ifdef PAWN_DEBUG
      int tmp = score;
       cout << ColorImage(side) << " pawn on " << FileImage(sq) << 
        RankImage(sq);
#endif
      int file = File(sq);
      int rank = Rank(sq,side);
      if (rank == 2) ++entr.pawns_on_2nd_rank;
      if (file < 4)
      {
        lside++;
      }
      else if (file > 5)
      {
        rside++;
      }
      if (SquareColor(sq) == White)
        entr.w_square_pawns++;
      else
        entr.b_square_pawns++;
      entr.open_files &= (byte)~(1<<(file-1));
      int backward = 0 /*FALSE*/;
      int passed = 0 /*FALSE*/;
      int weak = 0 /*FALSE*/;
      int doubled = 0 /*FALSE*/;
      int isolated;
      if (side == White)
      {
         entr.center_score += PawnCenterScores[sq];
      } 
      else
      {
         entr.center_score += PawnCenterScores[63-(int)sq];
      }
      Bitmap result;
      if (side == White)
         result = Bitmap::And(board.PawnBits(side),backwardW[sq]);
      else
         result = Bitmap::And(board.PawnBits(side),backwardB[sq]);
      backward = result.is_clear();
      if (side == White)
         result = Bitmap::And(board.PawnBits(oside),passedW[sq]);
      else
         result = Bitmap::And(board.PawnBits(oside),passedB[sq]);
      passed = result.is_clear();
      if (file == AFILE)
      {
         isolated = Bitmap::And(Bearing::file_mask[BFILE-1],
			       board.pawn_bits[side]).is_clear();
      }
      else if (file == HFILE)
      {
         isolated = Bitmap::And(Bearing::file_mask[GFILE-1],
			       board.pawn_bits[side]).is_clear();
      }
      else
      {
         isolated = Bitmap::And(Bearing::file_mask[file-2],
			       board.pawn_bits[side]).is_clear() &&
                   Bitmap::And(Bearing::file_mask[file],
			       board.pawn_bits[side]).is_clear();
      }
      Bitmap doubles = Bitmap::And(FileAttacksDown(board,sq),board.pawn_bits[side]);
      if (!doubles.is_clear()) {
#ifdef PAWN_DEBUG
	cout << " doubled";
#endif
         doubled++;
         score += DOUBLED_PAWNS;
         if (doubles.bit_count() > 1) score += TRIPLED_PAWNS;
         if (isolated) score += ISOLATED_DOUBLED;
      }
      if (isolated && !passed) {
        isolated_count++;
#ifdef PAWN_DEBUG
        cout << " isolated";
#endif
      }
      if (board[sq+incr] != enemy_pawn)
      {
        int defenders = pawn_attacks(board,sq,side).bit_count();
        if (defenders == 0)
        {
           Square sq2(sq + incr);
           int patcks = pawn_attacks(board,sq2,oside).bit_count() -
                        pawn_attacks(board,sq2,side).bit_count() + 1;
           if (patcks > 0) {
#ifdef PAWN_DEBUG
              cout << " weak";
#endif
              score += WEAK;
              weak++;
              if (Bitmap::And(board.pawn_bits[oside],
			      Bearing::file_mask[file-1]).is_clear()) {
                 score += WEAK_ON_OPEN_FILE;
                 entr.weak_files = (byte)(entr.weak_files | (byte)(1<<(file-1)));
#ifdef PAWN_DEBUG
		 cout << " weak/open";
#endif
	      }
           }
           if (patcks > 1) { // "extra weak" pawn 
              score += WEAK2;
#ifdef PAWN_DEBUG
  	      cout << " extra weak";
#endif

 	   }
        }
      }
      if (!isolated && !weak && !backward && !passed && rank > 2)
      {
          //score += UNSUPPORTED_PAWN;
          if (rank >4 && pawn_attacks(board,sq,side).is_clear()) {
	    score += ADV_UNSUPPORTED_PAWN;
#ifdef PAWN_DEBUG
	    cout << " unsupported" << endl;
#endif
	  }
      }
      if (passed)
      {
#ifdef PAWN_DEBUG
	cout << " passed";
#endif
         entr.passers[entr.passers_count++] = sq;
         ASSERT(entr.passers_count <= 8);
         if (file > entr.last_passer)
            entr.last_passer = file;
         if (file < entr.first_passer)
            entr.first_passer = file;
         score += PASSED_PAWN[rank-1];
         if (side == White)
         {
            Bitmap atcks(Bearing::pawn_attacks[sq][White]);
            atcks.And(board.pawn_bits[White]);
            if (!atcks.is_clear())
               entr.protected_pp = 1;
         } 
         else
         {
            Bitmap atcks(Bearing::pawn_attacks[sq][Black]);
            atcks.And(board.pawn_bits[Black]);
            if (!atcks.is_clear())
               entr.protected_pp = 1;
         }
         int j;
         for (j=0;j<entr.passers_count-1;j++)
         {
            Square psq = entr.passers[j];
            int psqfile = File(psq);
            if (psqfile == file-1 || psqfile == file+1)
            {
               int psqrank = Rank(psq,side);
               int rdiff = rank - psqrank;
               // score is based on the more advanced
               // of the 2 pawns
               if (rdiff == 0 || rdiff == 1)
               {
                  score += CONNECTED_PASSERS[rank-1];
#ifdef PAWN_DEBUG
               cout <<  " connected passer";
#endif   
               } 
               else if (rdiff == -1)
               {
                  score += CONNECTED_PASSERS[psqrank-1];
#ifdef PAWN_DEBUG
               cout <<  " connected passer";
#endif   
               }
               else
                  score += ADJACENT_PASSERS;
            }
         }
         if (file ==1 || file == 8)
            score += (RIM_PASSED_PAWN*(rank+1))/8;
         for (j=rank+1;j<9;j++)
         {
            Square sq2 = Square(file,j,side);
            Piece piece2 = board[sq2];
            if (TypeOfPiece(piece2) != EmptyPiece)
            {
               if (PieceColor(piece2) == side)
               {
                  if (TypeOfPiece(piece2) == Pawn)
                  {
                     // This pawn is passed but another
                     // pawn is ahead of it, i.e. it is
                     // doubled
                     score -= PASSED_PAWN[rank-1]/3;
                  }
               }
            }
         }
      }
      else if (Rank(side,sq) >= 5) {
         // A "hidden" passed pawn can advance and turn into a
         // passer, since it is part of a 2 vs. 1 pawn majority.
         // Crafty has similar scoring for this, but this code
         // is a little more general:
         int hidden = 0;
         if (File(sq) == HFILE &&
             Bitmap::And(all_pawns,Bearing::file_mask[HFILE-1]).is_clear()) {
             if (side == White) {
               Bitmap b = Bitmap::And(w_files_ahead[sq],board.all_occupied);
               b.And(board.pawn_bits[Black]);
               if (b.bit_count() == 1) {
                  Square sq2 = sq - 1;
                  if (board[sq2] == whitePawn) {
                     hidden = Bitmap::And(w_files_ahead[sq2],board.pawn_bits[Black]).is_clear();
                  } else if (board[sq2+8] == whitePawn) {
                     hidden = Bitmap::And(w_files_ahead[sq2+8],board.pawn_bits[Black]).is_clear();
                  } else if (board[sq2-8] == whitePawn) {
                     hidden = Bitmap::And(w_files_ahead[sq2-8],board.pawn_bits[Black]).is_clear();
                  }
               }
            } else {
               Bitmap b = Bitmap::And(b_files_ahead[sq],board.all_occupied);
               b.And(board.pawn_bits[White]);
               if (b.bit_count() == 1) {
                  Square sq2 = sq - 1;
                  if (board[sq2] == blackPawn) {
                     hidden = Bitmap::And(b_files_ahead[sq2],board.pawn_bits[Black]).is_clear();
                  } else if (board[sq2+8] == blackPawn) {
                     hidden = Bitmap::And(b_files_ahead[sq2+8],board.pawn_bits[Black]).is_clear();
                  } else if (board[sq2-8] == blackPawn) {
                     hidden = Bitmap::And(b_files_ahead[sq2-8],board.pawn_bits[Black]).is_clear();
                  }
               }
            }                        
         } else if (File(sq) == AFILE && 
             Bitmap::And(all_pawns,Bearing::file_mask[CFILE-1]).is_clear()) {
            if (side == White) {
               Bitmap b = Bitmap::And(w_files_ahead[sq],board.all_occupied);
               b.And(board.pawn_bits[Black]);
               if (b.bit_count() == 1) {
                  Square sq2 = sq + 1;
                  if (board[sq2] == whitePawn) {
                     hidden = Bitmap::And(w_files_ahead[sq2],board.pawn_bits[Black]).is_clear();
                  } else if (board[sq2+8] == whitePawn) {
                     hidden = Bitmap::And(w_files_ahead[sq2+8],board.pawn_bits[Black]).is_clear();
                  } else if (board[sq2-8] == whitePawn) {
                     hidden = Bitmap::And(w_files_ahead[sq2-8],board.pawn_bits[Black]).is_clear();
                  }
               }
            } else {
               Bitmap b = Bitmap::And(b_files_ahead[sq],board.all_occupied);
               b.And(board.pawn_bits[White]);
               if (b.bit_count() == 1) {
                  Square sq2 = sq + 1;
                  if (board[sq2] == blackPawn) {
                     hidden = Bitmap::And(b_files_ahead[sq2],board.pawn_bits[Black]).is_clear();
                  } else if (board[sq2+8] == blackPawn) {
                     hidden = Bitmap::And(b_files_ahead[sq2+8],board.pawn_bits[Black]).is_clear();
                  } else if (board[sq2-8] == blackPawn) {
                     hidden = Bitmap::And(b_files_ahead[sq2-8],board.pawn_bits[Black]).is_clear();
                  }
               }
            }
            if (hidden) {
               score += HIDDEN_PASSED_PAWN;
#ifdef PAWN_DEBUG
               cout <<  " hidden passer";
#endif   
            }
         }
      }
      int duo = 0;
      if (rank > 3 && file != 8 && board[sq+1] == our_pawn)
      {
#ifdef PAWN_DEBUG
	cout << " duo";
#endif
         score += PAWN_DUO;
	 duo++;
      }
      if (!passed && rank >= 4)
      {
         int space = SPACE*(rank-3);
         if (duo || !pawn_attacks(board,sq,side).is_clear())
            space += (SPACE*(rank-3))/2;
         if (rank >= 5 && board[sq+incr] == enemy_pawn)
         // cramping enemy pawn
            space += (SPACE*(rank-3))/2;
         score += space;
#ifdef PAWN_DEBUG
         cout << " space=" <<space; 
         total_space += space;
#endif         
      }
#ifdef PAWN_DEBUG
      cout << " total = " << score-tmp << endl;
#endif
   }
   ASSERT(isolated_count<8);
   score += ISOLATED_PAWN[isolated_count];
#ifdef PAWN_DEBUG
   cout << ColorImage(side) << " isolated pawn penalty = " << ISOLATED_PAWN[isolated_count] << endl;
   cout << ColorImage(side) << " pawn space score = " << total_space << endl;
#endif
   entr.score = score;
   return score;
}

// Calculate endgame-related info for a position and cache it
// in "kp_hash_entry"
//
static void calc_endgame(const Board &board, Pawn_Entry::PawnData *w_pawn_entry,
                         Pawn_Entry::PawnData *b_pawn_entry, King_Pawn_Entry *kp_hash_entry) {
   int k_pos = 0; Square kp = board.KingPos(White);
   Bitmap all_pawns(Bitmap::Or(board.pawn_bits[White],board.pawn_bits[Black]));
   if (!Bitmap::And(all_pawns,efgh_mask).is_clear() && !Bitmap::And(all_pawns,abcd_mask).is_clear())
      k_pos += KingEndgameScoresW[kp];
   else if (!Bitmap::And(all_pawns,efgh_mask).is_clear())
      k_pos += KingEndgameScoresWK[kp];
   else if (!Bitmap::And(all_pawns,abcd_mask).is_clear())
      k_pos += KingEndgameScoresWQ[kp];
   kp_hash_entry->white_king_position  = k_pos;
   k_pos = 0; kp = board.KingPos(Black);
   if (!Bitmap::And(all_pawns,efgh_mask).is_clear() && !Bitmap::And(all_pawns,abcd_mask).is_clear())
     k_pos += KingEndgameScoresB[kp];
   else if (!Bitmap::And(all_pawns,efgh_mask).is_clear())
     k_pos += KingEndgameScoresBK[kp];
   else if (!Bitmap::And(all_pawns,abcd_mask).is_clear())
     k_pos += KingEndgameScoresBQ[kp];
   kp_hash_entry->black_king_position = k_pos;
   kp_hash_entry->white_endgame_pawn_proximity = 0;
   kp_hash_entry->black_endgame_pawn_proximity = 0;
   if (Bitmap::And(all_pawns,left_side_mask[4]).is_clear() ||
       Bitmap::And(all_pawns,right_side_mask[3]).is_clear())
   {
      // bonus for king near pawns
      Bitmap it(all_pawns);
      Square sq;
      for (;(sq = it.first_one()) != InvalidSquare;it.clear(sq))
      {
         kp_hash_entry->white_endgame_pawn_proximity += (4-distance1(board.KingPos(White),sq));
         kp_hash_entry->black_endgame_pawn_proximity += (4-distance1(board.KingPos(Black),sq));
      }
   } 
   kp_hash_entry->w_uncatchable = kp_hash_entry->b_uncatchable = 0;
   int i ;
   for (i = 0; i < w_pawn_entry->passers_count; i++) {
      if (uncatchable(board,w_pawn_entry->passers[i],board.KingPos(White),board.KingPos(Black),White)) {
#ifdef PAWN_DEBUG
         cout << "White pawn on " <<  FileImage(w_pawn_entry->passers[i]) <<
              RankImage(w_pawn_entry->passers[i]) << " is uncatchable" << endl;
#endif
         kp_hash_entry->w_uncatchable |= (1<<i);
      }
   }
   for (i = 0; i < b_pawn_entry->passers_count; i++) {
      if (uncatchable(board,b_pawn_entry->passers[i],board.KingPos(Black),board.KingPos(White),Black)) {
#ifdef PAWN_DEBUG
         cout << "Black pawn on " << FileImage(b_pawn_entry->passers[i]) << RankImage(b_pawn_entry->passers[i]) << " is uncatchable" << endl;
#endif
         kp_hash_entry->b_uncatchable |= (1<<i);
      }
   }
}

// Calculate a "cover" score (pawn cover around the king, for
// king safety calculation). 
//
/*static*/ void calc_cover(const Board &board,Square kp,ColorType side,
                       int &cover, int &proximity)
{
   cover = proximity = 0;
   int kpfile = File(kp);
   const Piece our_pawn = MakePiece(Pawn,side);
   const Piece our_bishop = MakePiece(Bishop,side);
   const ColorType oside = OppositeColor(side);
   int kprank = Rank(kp,side);
   int kcover = 0, qcover = 0;
   int middle_open = 0;
   int k_defects =0, q_defects=0, middle_defects=0;

//   if (kpfile > 5)
   {
      Square pawnsq = Square(FFILE,2,side);
      if (board[pawnsq] != our_pawn)
      {
         Square pawnsq2 = Square(FFILE,3,side);
         if (board[pawnsq2] == our_pawn)
         {
            kcover += BP_COVER_MISSING;
         } 
         else {
            if (FileOpen(FFILE,side)) 
              kcover += FileOpen(FFILE,oside) ? BP_COVER_MISSING4 : BP_COVER_MISSING3;
            else
              kcover += BP_COVER_MISSING2;
            k_defects++;
         }
      }
      else if (board[Square(FFILE,3,side)] == our_pawn)
         kcover += DOUBLED_PAWN_NEAR_KING;
      pawnsq = (int)(pawnsq+1);
      if (board[pawnsq] != our_pawn)
      {
         Square pawnsq2 = Square(GFILE,3,side);
         if (board[pawnsq2] == our_pawn)
         {
            kcover += NP_COVER_MISSING;
            if (board[Square(GFILE,2,side)] == our_bishop)
              kcover += FIANCHETTO;
         } 
         else {
            if (FileOpen(GFILE,side))
              kcover += FileOpen(GFILE,oside) ? NP_COVER_MISSING4 : NP_COVER_MISSING3;
            else
              kcover += NP_COVER_MISSING2;
            k_defects++;
         }
      }
      else if (board[Square(GFILE,3,side)] == our_pawn)
         kcover += DOUBLED_PAWN_NEAR_KING;
      pawnsq = (int)(pawnsq+1);
      if (board[pawnsq] != our_pawn)
      {
         Square pawnsq2 = Square(HFILE,3,side);
         if (board[pawnsq2] == our_pawn)
            kcover += RP_COVER_MISSING;
         else {
            if (FileOpen(HFILE,side))
              kcover += FileOpen(HFILE,oside) ? RP_COVER_MISSING4 : RP_COVER_MISSING3;
            else
              kcover += RP_COVER_MISSING2;
            k_defects++;
         }
      }
      else if (board[Square(HFILE,3,side)] == our_pawn)
         kcover += DOUBLED_PAWN_NEAR_KING;
      if (kpfile==6 && FileOpen(EFILE,side)) {
         kcover += KING_NEAR_OPEN_FILE;
         if (FileOpen(EFILE,oside)) kcover += KING_NEAR_OPEN_FILE;
         k_defects++;
      }
   } 
//   else if (kpfile < 4)
   {
      Square pawnsq = Square(AFILE,2,side);
      if (board[pawnsq] != our_pawn)
      {
         Square pawnsq2 = Square(AFILE,3,side);
         if (board[pawnsq2] == our_pawn)
            qcover += RP_COVER_MISSING;
         else {
            if (FileOpen(AFILE,side))
              qcover += FileOpen(AFILE,oside) ? RP_COVER_MISSING4 : RP_COVER_MISSING3;
            else
              qcover += RP_COVER_MISSING2;
            q_defects++;
         }
      }
      else if (board[Square(AFILE,3,side)] == our_pawn)
         kcover += DOUBLED_PAWN_NEAR_KING;
      pawnsq = (int)(pawnsq+1);
      if (board[pawnsq] != our_pawn)
      {
         Square pawnsq2 = Square(BFILE,3,side);
         if (board[pawnsq2] == our_pawn)
         {
            qcover += NP_COVER_MISSING;
            if (board[Square(BFILE,2,side)] == our_bishop)
              qcover += FIANCHETTO;
         } 
         else {
            if (FileOpen(BFILE,side))
              qcover += FileOpen(BFILE,oside) ? NP_COVER_MISSING4 : NP_COVER_MISSING3;
            else
              qcover += NP_COVER_MISSING2;
            q_defects++;
         }
      }
      else if (board[Square(BFILE,3,side)] == our_pawn)
         kcover += DOUBLED_PAWN_NEAR_KING;
      pawnsq = (int)(pawnsq+1);
      if (board[pawnsq] != our_pawn)
      {
         Square pawnsq2 = Square(CFILE,3,side);
         if (board[pawnsq2] == our_pawn)
         {
            qcover += BP_COVER_MISSING;
         }
         else {
            if (FileOpen(CFILE,side))
              qcover += FileOpen(CFILE,oside) ? BP_COVER_MISSING4 : BP_COVER_MISSING3;
            else
              qcover += BP_COVER_MISSING2;
            q_defects++;
         }
      }
      else if (board[Square(CFILE,3,side)] == our_pawn)
         kcover += DOUBLED_PAWN_NEAR_KING;
      if (kpfile==3 && FileOpen(DFILE,side)) {
         qcover += KING_NEAR_OPEN_FILE;
         if (FileOpen(DFILE,oside)) qcover += KING_NEAR_OPEN_FILE;
      }
   }
   if (kpfile == 4 || kpfile == 5) {
      if (FileOpen(DFILE,side))
      {
         if (FileOpen(DFILE,oside))
         {
            middle_open += 2*MIDDLE_OPEN;
            middle_defects++;
         } 
         else
            middle_open += MIDDLE_OPEN;
      }
      if (FileOpen(EFILE,side))
      {
         if (FileOpen(EFILE,oside))
         {
            middle_open += 2*MIDDLE_OPEN;
            middle_defects++;
         } 
         else
            middle_open += MIDDLE_OPEN;
      }
      if (kpfile==5 && FileOpen(FFILE,side)) {
         middle_open += KING_NEAR_OPEN_FILE;
         if (FileOpen(FFILE,oside)) middle_open += KING_NEAR_OPEN_FILE;
         middle_defects++;
      }
      if (kpfile==4 && FileOpen(CFILE,side)) {
         middle_open += KING_NEAR_OPEN_FILE;
         if (FileOpen(CFILE,oside)) middle_open += KING_NEAR_OPEN_FILE;
         middle_defects++;
      }

   }
   if (kprank >= 2) {
      Bitmap b(Bearing::king_attacks[kp]);
      b.And(board.pawn_bits[side]);
      int pawns = b.bit_count();
      cover += 4*pawns-12;
      if (kprank >= 3) cover-=12;
   }
   if (kpfile > 5)
   {
      cover += kcover;
      if (k_defects>1) cover+=KS_DEFECTS[(k_defects-1)];
   }
   else if (kpfile < 4)
   {
      cover += qcover;
      if (q_defects>1) cover+=KS_DEFECTS[(q_defects-1)];
   }
   else 
   {
      cover += (kcover + qcover )/2 + middle_open;
      int d = (k_defects+q_defects)/2 + middle_defects;
      if (d>1) cover+=KS_DEFECTS[d-1];
      if (side == White)
      {
         Bitmap b(center_pawn_cover_w);
         b.And(board.pawn_bits[White]);
         cover += 2*((int)b.bit_count()-4);
      } 
      else
      {
         Bitmap b(center_pawn_cover_b);
         b.And(board.pawn_bits[Black]);
         cover += 2*((int)b.bit_count()-4);
      }
   }
  
   if (side == White)
   {
      Bitmap patcks(w_pawn_attacks[kpfile-1]);
      // check for enemy pawns near the king
      patcks.And(board.pawn_bits[Black]);
      if (!patcks.is_clear())
      {
         Bitmap p1(patcks);
         p1.And(Bearing::rank_mask[6]);
         proximity -= 11*p1.bit_count();
         p1 = patcks;
         p1.And(Bearing::rank_mask[5]);
         proximity -= 8*p1.bit_count();
         p1 = patcks;
         p1.And(Bearing::rank_mask[4]);
         proximity -= 5*p1.bit_count();
         p1 = patcks;
         p1.And(Bearing::rank_mask[3]);
         proximity -= 2*p1.bit_count();
      }
   } 
   else
   {
      // check for enemy pawns near the king
      Bitmap patcks(b_pawn_attacks[kpfile-1]);
      patcks.And(board.pawn_bits[White]);
      if (!patcks.is_clear())
      {
         Bitmap p1(patcks);
         p1.And(Bearing::rank_mask[1]);
         proximity -= 11*p1.bit_count();
         p1 = patcks;
         p1.And(Bearing::rank_mask[2]);
         proximity -= 8*p1.bit_count();
         p1 = patcks;
         p1.And(Bearing::rank_mask[3]);
         proximity -= 5*p1.bit_count();
         p1 = patcks;
         p1.And(Bearing::rank_mask[4]);
         proximity -= 2*p1.bit_count();
      }
   }
   proximity += proximity/2;
}

int Scoring::evalu8( const Board &board )
{
   endgame = in_endgame(board);
   if (endgame && is_draw(board))
      return 0;
   else
   {
#ifdef EVAL_DEBUG
     int mat_score = MaterialScore(board,board.Side());
     cout << "material score = " << mat_score << endl;
     return mat_score +
      positional_score(board,-Constants::BIG,Constants::BIG);
#else
      return MaterialScore(board,board.Side()) +
      positional_score(board,-Constants::BIG,Constants::BIG);
#endif
   }
}

int Scoring::is_draw(const Board &board)
{
   return (repetition_draw(board) || material_draw(board) ||
           is_theoretical_draw(board));
}        

int Scoring::is_legal_draw(const Board &board)
{
   return (repetition_draw(board) || material_draw(board));
}        

int Scoring::is_theoretical_draw(const Board &board)
{
   const Material &mat1 = board.getMaterial(White);
   const Material &mat2 = board.getMaterial(Black);
   if (mat1.value() > 2000+(BISHOP_VALUE*2) &&
       mat2.value() > 2000+(BISHOP_VALUE*2))
       return 0 /*FALSE*/;
   // Check for K + P vs rook pawn
   if ((unsigned)mat1.infobits() == Material::KP &&
       mat2.king_only())
   {
      if (!Bitmap::And(rook_pawn_mask,
                       board.pawn_bits[White]).is_clear())
      {
         Square kp = board.KingPos(White);
         Square kp2 = board.KingPos(Black);
         Square psq = (board.pawn_bits[White].first_one());
         if (Rank(kp,White) < Rank(kp2,White))
         {
            Square qsq = Square(File(psq),8,White);
            return !uncatchable(board,psq,kp,kp2,White) &&
            distance1(kp2,qsq) < distance1(kp,qsq);
         }
      }
   } 
   else if ((unsigned)mat2.infobits() == Material::KP &&
              mat1.king_only())
   {
      if (!Bitmap::And(rook_pawn_mask,
                       board.pawn_bits[Black]).is_clear())
      {
         Square kp = board.KingPos(Black);
         Square kp2 = board.KingPos(White);
         Square psq(board.pawn_bits[Black].first_one());
         if (Rank(kp,Black) < Rank(kp2,Black))
         {
            Square qsq = Square(File(psq),8,Black);
            return !uncatchable(board,psq,kp,kp2,Black) &&
            distance1(kp2,qsq) < distance1(kp,qsq);
         }
      }
   }
   // Check for wrong bishop + rook pawn vs king.  Not very common but
   // we don't want to get it wrong.
   else if (mat1.infobits() == Material::KBP && mat2.king_only())
   {
      if (KBPDrawW(board)) return 1;
   } 
   else if (mat2.infobits() == Material::KBP && mat1.king_only())
   {
      if (KBPDrawB(board)) return 1;
   }
   // check for KNN vs K
   if (((unsigned)mat1.infobits() == Material::KNN &&
        mat2.king_only()) ||
       ((unsigned)mat2.infobits() == Material::KNN &&
        mat1.king_only()))
      return 1 /*TRUE*/;

   return 0 /*FALSE*/;
}

static int score_endgame(const Board &board, ColorType side, Pawn_Entry::PawnData *hash_entry, King_Pawn_Entry *kp_hash_entry)
{
   const ColorType oside = OppositeColor(side);
   const Square kp = board.KingPos(side);
   const Square oppkp = board.KingPos(oside);
   int score = 0;
   const Material &my_mat = board.getMaterial(side);
   const Material &opp_mat = board.getMaterial(OppositeColor(side));
   if (board.Side() != side)
   {
      int x = opposition[kp][oppkp];
      if (x > 0)
         score += OPPOSITION*x;
   }
   // evaluate king position. Big bonus for centralizing king or (if
   // pawns are all on one side) being near pawns.
   // Similar to how Crafty does it.
   Bitmap all_pawns(board.pawn_bits[White]);
   all_pawns.Or(board.pawn_bits[Black]);
#ifdef EVAL_DEBUG
   int tmp = score;
   if (side == White) {
     score += (int)kp_hash_entry->white_king_position;
   }
   else {
     score += (int)kp_hash_entry->black_king_position;
   } 
   cout << "king position score " << ColorImage(side) << " = " << score-tmp << endl;
#else
   score += (side == White) ? (int)kp_hash_entry->white_king_position : 
            (int)kp_hash_entry->black_king_position;
#endif

   if ((my_mat.infobits() == Material::KNB) &&
       opp_mat.king_only())
   {
      // KBNK endgame, special case.
      const int opprank = Rank(oppkp,White);
      const int oppfile = File(oppkp);
      Bitmap bishops = board.bishop_bits[side];
      Square sq = bishops.first_one();
      // try to force king to a corner where mate is possible.
      Square evalsq;
      if (SquareColor(sq) == White)
      {
         evalsq = Square(9-opprank, oppfile, White);
      } 
      else
      {
         evalsq = oppkp;
      }
      score += KBNKScores[evalsq];
      // keep the kings close
      score += 10-distance1(board.KingPos(White),board.KingPos(Black));
      return score;
   } 
   else if ((my_mat.infobits() & 0xff000) != 0 &&
               (opp_mat.king_only() || 
                opp_mat.infobits() == Material::KN ||
                opp_mat.infobits() == Material::KB))
   {
      // bare K vs. R and/or Q, or Q vs 1 minor piece
      // "side" has the material advantage.
      // Encourage moving the opposing king to the edge.
      score += 2*(10-KingCenterScores[board.KingPos(oside)]);
      // Keep the kings close.
      score += 15-3*distance1(board.KingPos(White),board.KingPos(Black));
   } 
   else if ((my_mat.infobits() == Material::KP) &&
              (unsigned)opp_mat.king_only())
   {
      // KPK endgame
      Square passer = hash_entry->passers[0];
      int cant_catch = (side == White) ?
           (int)kp_hash_entry->w_uncatchable :
           (int)kp_hash_entry->b_uncatchable;
      if (cant_catch)
      {
         // The opposing king can't catch the pawn before
         // it queens
         score += (UNCATCHABLE_PAWN*Rank(passer,side))/7;
      } 
      else {
         // This should be a draw with correct play, so 
         // bring the score nearer zero:
         score -= 50; 
         int rank = Rank(passer,side);
         int file_dist = Util::Abs(File(passer) - File(kp));
         int rank_dist = Rank(kp,side)-rank;
         // strongly encourage putting the king ahead of the
         // pawn
         if (rank_dist > 0 && file_dist <= 1)
            score += KING_AHEAD_OF_PAWN;
         else if (rank_dist<=0) // king behind pawn
 	    score += 8-rank*2;
         //also encourage staying near pawn
         score += DISTANCE_FROM_PAWN*distance1(kp,passer);
      }
   } 
   else if (opp_mat.no_pieces() && opp_mat.pawn_count() <= 4 && hash_entry->passers_count)
   {
      // our opponent has no pieces, few pawns, and we have at least one
      // passed pawn
      int uncatchables = 0;
      int minfile = 9;
      int maxfile = 0;
      Square minfilesq, maxfilesq;
      // The following code takes account of "uncatchable pawns",
      // including special cases where there are two or more
      // pawns, not all of which can be stopped.
      for (int i = 0; i < hash_entry->passers_count; i++)
      {
         Square passer = hash_entry->passers[i];
         int uncatchable = (side == White) ?
           (int)(kp_hash_entry->w_uncatchable & (byte)(1<<i)) :
           (int)(kp_hash_entry->b_uncatchable & (byte)(1<<i)) ;
         if (uncatchable)
         {
            ++uncatchables;
            if (side == White)
            {
               if (Rank(passer,White) > white_passer_rank)
               {
                  white_passer_rank = Rank(passer,White);
                  white_passer_race_rank = white_passer_rank;
                  white_passer_sq = white_passer_race_sq = passer;
               }
            } 
            else
            {
               if (Rank(passer,Black) > black_passer_rank)
               {
                  black_passer_rank = Rank(passer,Black);
                  black_passer_race_rank = black_passer_rank;
                  black_passer_sq = black_passer_race_sq = passer;
               }
            }
         }
         if (File(passer) < minfile)
         {
            minfile = File(passer); minfilesq = passer;
         }
         if (File(passer) > maxfile)
         {
            maxfile = File(passer); maxfilesq = passer;
         }
      }
      if (!uncatchables && hash_entry->passers_count > 1)
      {
         int d21 = distance1(oppkp,minfilesq);
         int d22 = distance1(oppkp,maxfilesq);
         int x1 = d21 + Rank(maxfilesq,side) -1;
         int x2 = d22 + Rank(minfilesq,side) -1;
         int y = 0;
         if (board.Side() != side) y = 1;
         if (x1+y >= 8 && x2+y >= 8)
         {
            uncatchables++;
            if (side == White)
            {
               int rank = Util::Max(Rank(maxfilesq,White),
                                    Rank(minfilesq,White));
               if (rank > white_passer_rank)
               {
                  white_passer_rank = rank;
                  if (rank == Rank(maxfilesq,White))
                     white_passer_sq = maxfilesq;
                  else
                     white_passer_sq = minfilesq;
               }
            } 
            else
            {
               int rank = Util::Max(Rank(maxfilesq,Black),
                                    Rank(minfilesq,Black));
               if (rank > black_passer_rank)
               {
                  black_passer_rank = rank;
                  if (rank == Rank(maxfilesq,Black))
                     black_passer_sq = maxfilesq;
                  else
                     black_passer_sq = minfilesq;
               }
            }
         } 
	 else if (x2+y < 8 && uncatchable(board,
                                          Square(File(minfilesq),
                                          x2+y,
                                          side),
                                          kp,
                                          maxfilesq,side))
         {
            uncatchables++;
            if (side == White)
            {
               int rank = Util::Max(Rank(maxfilesq,White),
                                    Rank(minfilesq,White));
               if (rank > white_passer_rank)
               {
                  if (rank == Rank(maxfilesq,White))
                     white_passer_sq = maxfilesq;
                  else
                     white_passer_sq = minfilesq;
                  white_passer_rank = rank;
               }
            } 
  	    else
            {
               int rank = Util::Max(Rank(maxfilesq,Black),
                                    Rank(minfilesq,Black));
               if (rank > black_passer_rank)
               {
                  if (rank == Rank(maxfilesq,Black))
                     black_passer_sq = maxfilesq;
                  else
                     black_passer_sq = minfilesq;
                  black_passer_rank = rank;
               }
            }
         } 
	 else if (x1+y < 8 && uncatchable(board,
                                          Square(File(maxfilesq),
                                          x1+y,
                                          side),
                                          kp,
                                          minfilesq,side))
         {
            uncatchables++;
            if (side == White)
            {
               int rank = Util::Max(Rank(maxfilesq,White),
                                    Rank(minfilesq,White));
               if (rank > white_passer_rank)
               {
                  if (rank == Rank(maxfilesq,White))
                     white_passer_sq = maxfilesq;
                  else
                     white_passer_sq = minfilesq;
                  white_passer_rank = rank;
               }
            } 
            else
            {
               int rank = Util::Max(Rank(maxfilesq,Black),
                                    Rank(minfilesq,Black));
               if (rank > black_passer_rank)
               {
                  if (rank == Rank(maxfilesq,Black))
                     black_passer_sq = maxfilesq;
                  else
                     black_passer_sq = minfilesq;
                  black_passer_rank = rank;
               }
            }
         }
      }
      // give a large bonus for uncatchable pawns only if our
      // opponent has zero pawns; otherwise it is not sure we
      // can win
      int bonus = opp_mat.pawn_count() == 0 ? UNCATCHABLE_PAWN :
        8*(5-opp_mat.pawn_count());
      if (uncatchables) {
         score += bonus*Util::Min(uncatchables,2);
      }
      else  {
         score += (side == White) ?
            kp_hash_entry->white_endgame_pawn_proximity :
            kp_hash_entry->black_endgame_pawn_proximity;
      }
   } 
   else
   {
      // This code is VERY crude .. it helps (a little) in king and
      // pawn endgames, but is pretty useless otherwise ... Lots more
      // work to do here.
      score += (side == White) ?
            kp_hash_entry->white_endgame_pawn_proximity :
            kp_hash_entry->black_endgame_pawn_proximity;

      // give a bonus for unstoppable passers depending on the
      // material our opponent has.  (Our opponent has
      // at least one piece).
      int xtramat = opp_mat.value() - 2000;
      if (xtramat < 2000)
      {
         int uncatchables = 0;
         for (int i = 0; i < hash_entry->passers_count; i++)
         {
            Square pawnpos = hash_entry->passers[i];
            int cant_catch = (side == White) ?
              (int)(kp_hash_entry->w_uncatchable & (byte)(1<<i)) :
              (int)(kp_hash_entry->b_uncatchable & (byte)(1<<i)) ;
            if (cant_catch) {
               uncatchables++;
            } 
            else
            {
               // bonus for king near passer
               int rank = Rank(pawnpos,side);
               if (rank>=5) {
                    // advanced passer protected by King
                    score += 2*Rank(pawnpos,side);
               }
            }   
         }
         if (uncatchables)
         {
            int incr = UNCATCHABLE_PAWN2[board.OppositeSide() == White ?
                w_material_level : b_material_level];
            score += incr;
            if (uncatchables > 1)
               score += incr/2;
            if (uncatchables >2)
               score += incr*(uncatchables-2)/4;
         }
      }
   }
   return score;
}

// Calculates a net king saftety score from the perspective
// of the White side
//
static int king_safety(const Board &board,King_Pawn_Entry *hash_entry)
{
   int score = 0;
   int w_proximity = 0, b_proximity = 0;
   int w_cover = hash_entry->white_cover;
   int b_cover = hash_entry->black_cover;
   int w_count =  -hash_entry->white_pawn_proximity/6, b_count = -hash_entry->black_pawn_proximity/6;
   int w_king_pos = 0, b_king_pos = 0;
   
   Square kp = board.KingPos(White);
   Square whiteQ = InvalidSquare; 
   Square blackQ = InvalidSquare;

   int kprank = Rank(kp,White);
   int kpfile = File(kp);
   if (b_material_level > 2)
   {
      Bitmap result(Bitmap::And(king_proximity[kp],board.occupied[Black]));
      result.andNot(board.pawn_bits[Black]);
      result.andNot(board.knight_bits[Black]);
      w_count += result.bit_count();

      if (kprank >= 2) {
         w_king_pos += KING_OFF_BACK_RANK;
         if (kprank >= 3) w_king_pos += (3*KING_OFF_BACK_RANK)/2;
      }
      if (kprank < 4)
      {
         if (board[E3] == whitePawn && kpfile > EFILE) {
            if (Bitmap::And(b_stonewall,board.pawn_bits[Black]) == b_stonewall)
               w_proximity += STONEWALL;
            else if (board[F2] == blackPawn && (board[D4] == blackPawn || (TypeOfPiece(board[E4])==Pawn)))
               w_proximity += PAWN_WALL;
         }
         if (kpfile >= EFILE) {
            if (kpfile > EFILE) {
              if (!Bitmap::And(DiagAtcksA8Upper(board,H3),board.bishopsqueens[Black]).is_clear())
                w_count++;
              if (!Bitmap::And(DiagAtcksA8Upper(board,H2),board.bishopsqueens[Black]).is_clear())
                w_count++;
            }
            if (!Bitmap::And(DiagAtcksA8Upper(board,G2),board.bishopsqueens[Black]).is_clear())
              w_count++;
            if (kprank <= 2 && board[H3] == blackBishop && kpfile > FFILE) w_count++;
         }
         else {
            if (kpfile < DFILE) {
              if (!Bitmap::And(DiagAtcksA1Upper(board,A3),board.bishopsqueens[Black]).is_clear())
                 w_count++;
               if (!Bitmap::And(DiagAtcksA1Upper(board,A2),board.bishopsqueens[Black]).is_clear())
                 w_count++;
            }
            if (!Bitmap::And(DiagAtcksA1Upper(board,B2),board.bishopsqueens[Black]).is_clear())
              w_count++;
            if (kprank <= 2 && board[A3] == blackBishop && kpfile < CFILE) w_count++;
         }
      }
   }
   Bitmap rooksqueens(board.rooksqueens[Black]);
   Square sq;
   for (;(sq = rooksqueens.first_one()) != InvalidSquare;rooksqueens.clear(sq))
   {
     int qbonus = 0;
     int rank = Rank(sq,Black);
     if ((rank == 7 || rank == 8)&& kprank<=3)
     {
         w_count++;
         if (TypeOfPiece(board[sq]) == Queen) qbonus++;
         if (rank == 7 && TypeOfPiece(board[sq])==Rook) {
            score -= ROOK_ON_7TH_RANK;
            if (!Bitmap::And(RankAttacks(board,sq),board.rooksqueens[Black]).is_clear())
               w_count++;
        }
     }
     if (king_file_proximity[kpfile-1].is_set(sq))
     {
        w_count++;
        if (TypeOfPiece(board[sq]) == Queen) qbonus++;
        int sq2 = sq; 
        int d = 0;
        while (sq2 + 8 < 64) {
	      sq2 += 8; 
          if (board[sq2] == blackRook || 
              board[sq2] == blackQueen)
          {
             ++d;
             continue;
          }
          if (!d && Rank(sq2,White)<=kprank+1 && (board[sq2] == EmptyPiece ||
              PieceColor(board[sq2]) == Black || 
              Bitmap::And(Bearing::pawn_attacks[sq2][White],board.pawn_bits[White]).is_clear())) {
            w_count++;
            break;
          }
          else if (board[sq2] != EmptyPiece)
          { 
             break;
          }
        }
     }
     if (TypeOfPiece(board[sq]) == (int)Queen)
     {
       blackQ = sq;
       w_count += Util::Max(0,KING_QUEEN_DISTANCE[distance1(sq,kp)]-qbonus);
     }
   }
   Bitmap knight_attacks(king_knight_proximity[kp]);
   knight_attacks.And(board.knight_bits[Black]);
   w_count += knight_attacks.bit_count();

   kp = board.KingPos(Black);
   kprank = Rank(kp,Black);
   kpfile = File(kp);
   if (w_material_level > 2)
   {
      Bitmap result(Bitmap::And(king_proximity[kp],board.occupied[White]));
      result.andNot(board.pawn_bits[White]);
      result.andNot(board.knight_bits[White]);
      b_count += result.bit_count();

      if (kprank >= 2) {
         b_king_pos += KING_OFF_BACK_RANK;
         if (kprank >= 3) b_king_pos += (3*KING_OFF_BACK_RANK)/2;
      }
      if (kprank < 4)
      {
         
         if (board[E6] == blackPawn && kpfile > EFILE) {
              if ( Bitmap::And(w_stonewall,board.pawn_bits[White]) == w_stonewall)
 			     b_proximity += STONEWALL;
			  else if (board[F7] == blackPawn && (board[D5] == blackPawn || (TypeOfPiece(board[E5])==Pawn)))
				 b_proximity += PAWN_WALL;
         }
         if (kpfile >= EFILE) {
            if (kpfile > EFILE) {
              if (!Bitmap::And(DiagAtcksA1Lower(board,H6),board.bishopsqueens[White]).is_clear())
                 b_count++;
              if (!Bitmap::And(DiagAtcksA1Lower(board,H7),board.bishopsqueens[White]).is_clear())
                 b_count++;
            }
            if (!Bitmap::And(DiagAtcksA1Lower(board,G7),board.bishopsqueens[White]).is_clear())
              b_count++;
            if (kprank <= 2 && board[H6] == whiteBishop && kpfile > FFILE) b_count++;
         }
         else {
            if (kpfile < DFILE) {
              if (!Bitmap::And(DiagAtcksA8Lower(board,A6),board.bishopsqueens[White]).is_clear())
                 b_count++;
               if (!Bitmap::And(DiagAtcksA8Lower(board,A7),board.bishopsqueens[White]).is_clear())
                 b_count++;
            }
            if (!Bitmap::And(DiagAtcksA8Lower(board,B7),board.bishopsqueens[White]).is_clear())
              b_count++;
            if (kprank <= 2 && board[A6] == whiteBishop && kpfile < CFILE) b_count++;
         }
      }
   }
   rooksqueens = (board.rooksqueens[White]);
   for (;(sq = rooksqueens.first_one()) != InvalidSquare;rooksqueens.clear(sq))
   {
     int qbonus = 0;
     int rank = Rank(sq,White);
     if ((rank == 7 || rank == 8)&& kprank<=3)
     {
        b_count++;
        if (TypeOfPiece(board[sq]) == Queen) qbonus++;
        if (rank == 7 && TypeOfPiece(board[sq])==Rook) {
            score += ROOK_ON_7TH_RANK;
            if (!Bitmap::And(RankAttacks(board,sq),board.rooksqueens[White]).is_clear())
               b_count++;
        }
     }
     if (king_file_proximity[kpfile-1].is_set(sq))
     {
       b_count++;
       if (TypeOfPiece(board[sq]) == Queen) qbonus++;
       int sq2 = sq;
       int d = 0;
       while (sq2 - 8 >= 0) {
          sq2 -= 8; 
          if (board[sq2] == whiteRook || 
              board[sq2] == whiteQueen)
          {
             ++d;
             continue;
          }
          if (!d && Rank(sq2,Black)<=kprank+1 && (board[sq2] == EmptyPiece ||
              PieceColor(board[sq2]) == White ||
              Bitmap::And(Bearing::pawn_attacks[sq2][Black],board.pawn_bits[Black]).is_clear())) {
            b_count++;
            break;
          }
          else if (board[sq2] != EmptyPiece)
          { 
             break;
          }
        }
     }
     if (TypeOfPiece(board[sq]) == (int)Queen)
     {
       whiteQ = sq;
       b_count += Util::Max(0,KING_QUEEN_DISTANCE[distance1(sq,kp)]-qbonus);
     }
   }  
   knight_attacks = king_knight_proximity[kp];
   knight_attacks.And(board.knight_bits[White]);
   b_count += knight_attacks.bit_count();

#ifdef EVAL_DEBUG
   int tmp=score;
#endif
   if (b_material_level >= 2)
   {
      int p = -PROXIMITY_SCALE[w_count]+w_proximity;
      w_cover -= w_cover/3;
      if (b_material_level<=6) w_cover -= w_cover/4;
      int boost = -(w_cover*p)/40;
      int factor  = KS_MATERIAL_SCALE[b_material_level];
      int d = 0;
      if (w_count > b_count+1 && !IsInvalid(whiteQ)) {
         int d1 = distance1(whiteQ,board.KingPos(White));
         if (d1>2)
           d = -w_count*(d1-2);
#ifdef  EVAL_DEBUG
         if (d != 0)
            cout << "wandering queen (White), penalty = " << d << endl;
#endif
      }
      score += Util::Max(-180,((w_cover + p + boost + w_king_pos + d)*factor)/16);
#ifdef EVAL_DEBUG
      cout << "king safety(w) count=" << w_count << " proximity=" << p <<
      " cover=" << w_cover << " boost=" << boost << " mlevel=" <<
      b_material_level << " total=" << (score-tmp) << endl;
#endif
   }
#ifdef EVAL_DEBUG
   tmp = score;
#endif
   if (w_material_level >= 2)
   {
      int p = -PROXIMITY_SCALE[b_count]+b_proximity;
      b_cover -= b_cover/3;
      if (w_material_level<=6) b_cover -= b_cover/4;
      int boost = -(b_cover*p)/40;
      int factor  = KS_MATERIAL_SCALE[w_material_level];
      int d = 0;
      if (b_count > w_count+1 && !IsInvalid(blackQ)) {
         int d1 = distance1(blackQ,board.KingPos(Black));
         if (d1>2)
         {
           d = -b_count*(d1-2);
#ifdef  EVAL_DEBUG
           if (d != 0)
              cout << "wandering queen (Black), penalty = " << d << endl;
#endif
         }
      }
      score -= Util::Max(-180,((b_cover + p + boost + b_king_pos + d)*factor)/16);
#ifdef EVAL_DEBUG
      cout << "king safety(b) count=" << b_count << " proximity=" << p <<
      " cover=" << b_cover << " boost=" << boost << " mlevel=" <<
      w_material_level << " total=" << (score-tmp) << endl;
#endif
   }

#ifdef EVAL_DEBUG
   cout << "net king safety score = " << score << endl;
#endif   

   return score;
}



#define PAWN_RACE_WINNER(rank) (UNCATCHABLE_PAWN)

int Scoring::positional_score( const Board &board, int alpha, int beta )
// returns a positional score. Note: scores are from
// the perspective of the White side.
{
   int endgame = in_endgame(board);
   int score = 0;
   int i;
#ifdef LAZY_DEBUG
   int alpha_cut = 0, beta_cut =0;
#endif

   // Calculate material level (0-12 for each side)
   w_material_level = Util::Min(12,board.getMaterial(White).material_level());
   b_material_level = Util::Min(12,board.getMaterial(Black).material_level());

   // Query the pawn hash table to see if we have found a position
   // before that has the same pawn structure.  If so, we can skip
   // some of the scoring calculations.

   pawn_hash_entry = &pawn_hash_table[quickmod(board.pawn_hash(),PAWN_HASH_SIZE)];
   if (pawn_hash_entry->hash_code() != board.pawn_hash())
   {
      // pawn position not found, calculate the data we need
      pawn_structure_score(board,White,pawn_hash_entry->w_pawn_entry);
      pawn_structure_score(board,Black,pawn_hash_entry->b_pawn_entry);
      eval_outside_pp(board,pawn_hash_entry);
      pawn_hash_entry->hc = board.pawn_hash();
   }
   hash_entry_w = &pawn_hash_entry->w_pawn_entry;
   hash_entry_b = &pawn_hash_entry->b_pawn_entry;
#ifdef EVAL_DEBUG
   cout << "white pawn score = " << hash_entry_w->score << endl;
   cout << "black pawn score = " << hash_entry_b->score << endl;
#endif
   score += hash_entry_w->score - hash_entry_b->score;   
   if (!endgame) {
      score += hash_entry_w->center_score - hash_entry_b->center_score;
   }
#ifdef EVAL_DEBUG
   int tmp = score;
#endif
   // outside passed pawn scoring, based on cached pawn data
   if (pawn_hash_entry->w_pawn_entry.outside && !pawn_hash_entry->b_pawn_entry.outside)
   {
      score += OUTSIDE_PP[ b_material_level];
   } 
   else if (pawn_hash_entry->b_pawn_entry.outside && !pawn_hash_entry->w_pawn_entry.outside)
   {
      score -= OUTSIDE_PP[w_material_level];
   }
   // if there is no outside pawn, then protected pawns are very strong
/**   if (pawn_hash_entry->w_pawn_entry.outside == 0 && pawn_hash_entry->b_pawn_entry.outside == 0)
   {
      if (pawn_hash_entry->w_pawn_entry.protected_pp && !pawn_hash_entry->b_pawn_entry.protected_pp)
      {
         score += OUTSIDE_PP[b_material_level];
      }
      if (pawn_hash_entry->b_pawn_entry.protected_pp && !pawn_hash_entry->w_pawn_entry.protected_pp)
      {
         score -= OUTSIDE_PP[w_material_level];
      }
   }**/
#ifdef EVAL_DEBUG
   cout << "outside pawn score = " << tmp-score  << endl;
#endif
   if (board[A7] == whiteBishop)
   {
      if (board[B6] == blackPawn &&
          board.any_attacks(B6,Black))
      {
         score += BISHOP_TRAPPED;
      }
   }
   if (board[H7] == whiteBishop)
   {
      if (board[G6] == blackPawn &&
          board.any_attacks(G6,Black))
      {
         score += BISHOP_TRAPPED;
      }
   }
   if (board[A2] == blackBishop)
   {
      if (board[B3] == whitePawn &&
          board.any_attacks(B3,White))
      {
         score -= BISHOP_TRAPPED;
      }
   }
   if (board[H2] == blackBishop)
   {
      if (board[G3] == whitePawn &&
          board.any_attacks(G3,White))
      {
         score -= BISHOP_TRAPPED;
      }
   }

   King_Pawn_Entry *kp_hash_entry;
   hash_t kp_hash = BoardHash::kingPawnHash(board);
   kp_hash_entry = &king_pawn_hash_table[quickmod(kp_hash,KING_PAWN_HASH_SIZE)];
   if (kp_hash_entry->hc != kp_hash)
   {
      kp_hash_entry->cover_valid = 0;
      if (endgame) {
        calc_endgame(board,hash_entry_w,hash_entry_b,kp_hash_entry);
      }
      kp_hash_entry->endgame = endgame;
      kp_hash_entry->hc = kp_hash;
   }
   if (endgame)
   {
      if (!kp_hash_entry->endgame) {
         kp_hash_entry->endgame = 1;
         calc_endgame(board,hash_entry_w,hash_entry_b,kp_hash_entry);
      }
#if defined(EVAL_DEBUG) || defined(LAZY_DEBUG)
      int tmp = score;
#endif
      white_passer_rank = black_passer_rank = 
      white_passer_race_rank = black_passer_race_rank = 0;
      white_passer_race_sq = black_passer_race_sq = InvalidSquare;
      white_passer_sq = black_passer_sq = InvalidSquare;
#ifdef EVAL_DEBUG
      int endw = score_endgame(board,White,hash_entry_w,kp_hash_entry);
      int endb = score_endgame(board,Black,hash_entry_b,kp_hash_entry);
      cout << "endgame score(w) = " << endw << endl;
      cout << "endgame score(b) = " << endb << endl;
      score += (endw-endb);
#else
      score += score_endgame(board,White,hash_entry_w,kp_hash_entry) -
               score_endgame(board,Black,hash_entry_b,kp_hash_entry);
#endif
      Bitmap pawns(board.pawn_bits[White]);
      pawns.Or(board.pawn_bits[Black]);
      if (Bitmap::And(pawns,right_mask).data &&
          Bitmap::And(pawns,left_mask).data) {
 	     const Material &w_mat = board.getMaterial(White);
	     const Material &b_mat = board.getMaterial(Black);
         if (b_material_level <= 5 &&
	     w_mat.has_bishop() && !b_mat.has_bishop() && b_mat.has_knight())
	 {
            score += BISHOP_ENDGAME_BONUS[b_material_level];
#ifdef EVAL_DEBUG
            cout << "bishop endgame bonus(w) = " << BISHOP_ENDGAME_BONUS[b_material_level] << endl;
#endif
	 }
     else if (w_material_level <= 5 &&
	    !w_mat.has_bishop() && b_mat.has_bishop() && w_mat.has_knight())
	 {
            score -= BISHOP_ENDGAME_BONUS[w_material_level];
#ifdef EVAL_DEBUG
            cout << "bishop endgame bonus(b) = " << -BISHOP_ENDGAME_BONUS[w_material_level] << endl;
#endif

	 }
      }

#if defined(EVAL_DEBUG)
      tmp = score;
#endif
      // The following pawn race code is effective only when there is only a single
      // passer per side
      if ((w_material_level == 0 || b_material_level == 0) &&
           black_passer_race_rank > 0  && white_passer_race_rank > 0)
      {
         ASSERT(black_passer_race_sq != InvalidSquare);
         ASSERT(white_passer_race_sq != InvalidSquare);
         // evaluate pawn race
         // Decrement rank of passer if the king obstructs it
         if (File(board.KingPos(Black))==File(black_passer_race_sq) &&
             Rank(board.KingPos(Black),Black) > black_passer_race_rank)
            --black_passer_race_rank;
         if (File(board.KingPos(White))==File(white_passer_race_sq) &&
             Rank(board.KingPos(White),White) > white_passer_race_rank)
            --white_passer_race_rank;
         if (black_passer_race_rank >= white_passer_race_rank+2 ||
             (black_passer_race_rank == white_passer_race_rank+1 && board.Side() == Black))
            score -= PAWN_RACE_WINNER(black_passer_race_rank);
         else if (white_passer_race_rank >= black_passer_race_rank+2 ||
                  (white_passer_race_rank == black_passer_race_rank+1 && board.Side() == White))
            score += PAWN_RACE_WINNER(white_passer_race_rank);
         else if (white_passer_race_rank == black_passer_race_rank)
         {
            Square white_queen_square = Square(
                                              File(white_passer_race_sq),8,White);
            Square black_queen_square = Square(
                                              File(black_passer_race_sq),8,Black);
            if (board.Side() == White)
            {
               if (white_passer_race_rank == black_passer_race_rank)
               {
                  // White will queen first
                  // See if White will check the black king or
                  // if the White queen will stop the Black
                  // passer.
                  if (board.Clear(white_queen_square,
                                  board.KingPos(Black)) ||
                      (board.Clear(black_queen_square,
                                   board.KingPos(White)) &&
                       distance1(board.KingPos(Black),black_queen_square)>1
                      ))
                  {
                     score += PAWN_RACE_WINNER(white_passer_race_rank);
                  }
                  // See if Black will queen with check
                  else if (board.Clear(black_queen_square,
                                       board.KingPos(White)))
                  {
                     score -= PAWN_RACE_WINNER(black_passer_race_rank);
                  }
               }
            } 
            else
            {
               if (black_passer_race_rank == white_passer_race_rank)
               {
                  // Black will queen first.
                  // See if Black will check the White king or
                  // if the Black queen will stop the White
                  // passer.
                  if (board.Clear(black_queen_square,
                                  board.KingPos(White)) ||
                      (board.Clear(white_queen_square,
                                   board.KingPos(Black)) &&
                       distance1(board.KingPos(White),white_queen_square)>1
                      ))
                  {
                     score -= PAWN_RACE_WINNER(black_passer_race_rank);
                  }
                  // See if White will queen with check
                  else if (board.Clear(white_queen_square,
                                       board.KingPos(Black)))
                  {
                     score -= PAWN_RACE_WINNER(white_passer_race_rank);
                  }
               }
            }

         }
      }
#ifdef EVAL_DEBUG
      cout << "pawn race score = " << score - tmp << endl;
#endif
   } 
   // interaction of passed pawns and pieces

   Square wk = board.KingPos(White);
   
   for (i = 0; i < hash_entry_w->passers_count; i++)
   {
      Square sq = hash_entry_w->passers[i];
      ASSERT(OnBoard(sq));
      int file = File(sq);
      int rank = Rank(sq,White);
      // score passed pawns less if large material.
      // But not if opponent has bare king.
      if (!board.getMaterial(Black).king_only())
         score += PASSER_MATERIAL_BONUS[b_material_level]-6;
      int pp_block = 0;
      Square sq2 = sq-8;
      if (!IsEmptyPiece(board[sq2]) &&  (board[sq2] == blackKnight ||
          board[sq2] == blackBishop))
      {
          score -= PASSED_PAWN_BLOCK[rank-1];
          ++pp_block;
      }
      if (!board.rooksqueens[White].is_clear()) {
         if (!Bitmap::And(FileAttacksDown(board,sq),board.rooksqueens[White]).is_clear()) {
            int bonus = ROOK_BEHIND_PP;
            if (pp_block) bonus /= 2;
            score += bonus;
         }
         // give penalty if rook is ahead of a passer
         if (!Bitmap::And(FileAttacksUp(board,sq),board.rook_bits[White]).is_clear())
   	        score += ROOK_AHEAD_OF_PP;
         // also give penalty if protecting a passer with a rook from the side.
         if (!Bitmap::And(RankAttacks(board,sq),board.rook_bits[White]).is_clear())
   	        score += ROOK_AHEAD_OF_PP;
      }
      if (rank == 6 && (File(wk) == file-1 || File(wk) == file+1) &&
          Rank(wk,White) >= rank)
        score += SUPPORTED_PASSER6;
      else if (rank == 7 && (File(wk) == file-1 || File(wk) == file+1) &&
          Rank(wk,White) >= rank)
        score += SUPPORTED_PASSER7;
   }
   Square bk = board.KingPos(Black);
   for (i = 0; i < hash_entry_b->passers_count; i++)
   {
      Square sq = hash_entry_b->passers[i];
      ASSERT(OnBoard(sq));
      int file = File(sq);
      int rank = Rank(sq,Black);
      // score passed pawns less if large material
      if (!board.getMaterial(White).king_only())
         score -= PASSER_MATERIAL_BONUS[w_material_level]-6;
      int pp_block = 0;
      Square sq2 = sq+8;
      if (!IsEmptyPiece(board[sq2]) &&  (board[sq2] == whiteKnight ||
          board[sq2] == whiteBishop)) {
         score += PASSED_PAWN_BLOCK[rank-1];
         pp_block++;
      }
      if (!board.rooksqueens[Black].is_clear()) {
         if (!Bitmap::And(FileAttacksUp(board,sq),board.rooksqueens[Black]).is_clear()) {
            int bonus = ROOK_BEHIND_PP;
            if (pp_block) bonus /= 2;
            score -= bonus;
         }
         // give penalty if rook is ahead of a passer
         if (!Bitmap::And(FileAttacksDown(board,sq),board.rook_bits[Black]).is_clear())
   	        score -= ROOK_AHEAD_OF_PP;
         // also give penalty if protecting a passer with a rook from the side.
         if (!Bitmap::And(RankAttacks(board,sq),board.rook_bits[Black]).is_clear())
   	        score -= ROOK_AHEAD_OF_PP;
      }
      if (rank == 6 && (File(bk) == file-1 || File(bk) == file+1) &&
          Rank(bk,White) >= rank)
        score -= SUPPORTED_PASSER6;
      else if (rank == 7 && (File(bk) == file-1 || File(bk) == file+1) &&
          Rank(bk,White) >= rank)
        score -= SUPPORTED_PASSER7;
   }

   // if very low material only endgame terms matter, so can
   // exit here
//   if (endgame && w_material_level <= 2 && b_material_level <= 2)
//      return board.Side() == White ? score : -score;

   if (!endgame)
   {
      /** castling **/
      switch (board.CastleStatus(White))
      {
      case Board::CanCastleEitherSide:
         score += CAN_CASTLE; 
         break;
      case Board::CanCastleKSide:
         score += CANT_CASTLE_Q_SIDE; 
         break;
      case Board::CanCastleQSide:
         score += CANT_CASTLE_K_SIDE; 
         break;
      case Board::CastledKSide:
         score += KCASTLE; break;
      case Board::CastledQSide:
         score += QCASTLE; break;
      case Board::CantCastleEitherSide:
         score += CANT_CASTLE; break;
      }
      switch (board.CastleStatus(Black))
      {
      case Board::CanCastleEitherSide:
         score -= CAN_CASTLE; 
         break;
      case Board::CanCastleKSide:
         score -= CANT_CASTLE_Q_SIDE; 
         break;
      case Board::CanCastleQSide:
         score -= CANT_CASTLE_K_SIDE; 
         break;
      case Board::CastledKSide:
         score -= KCASTLE; break;
      case Board::CastledQSide:
         score -= QCASTLE; break;
      case Board::CantCastleEitherSide:
         score -= CANT_CASTLE; break;
      }
   }
   // If we're very far away from the bounds, we can exit without
   // doing the king safety calculation.
   // But be conservative: king safety scores can be large
   if (w_material_level > 2 || b_material_level > 2)
   {
#ifdef USE_LAZY_EVAL
   int lazy0 = PAWN_VALUE*4;
#ifdef LAZY_DEBUG
   int lazy_score0 = score;
#endif
   if (w_material_level <= 6 && b_material_level <= 6)
       lazy0 = PAWN_VALUE*3;
   int adjust1 = 0, adjust2 = 0;
   if (board.Side() == White) {
     if (kp_hash_entry->white_cover > -10)
        adjust1 = 100;
    if (kp_hash_entry->black_cover > -10)
        adjust2 = 100;
   }
   else {
     if (kp_hash_entry->black_cover > -10)
       adjust1 = 100;
     if (kp_hash_entry->white_cover > -10)
        adjust2 = 100;
   }
#ifdef LAZY_DEBUG
   if (board.Side() == White) {
      if (lazy0-adjust2+score < alpha && board.getMaterial(White).queen_count() < 2)
    	 alpha_cut++;
      if (score+adjust1-lazy0 > beta && board.getMaterial(Black).queen_count() < 2)
    	 beta_cut++;
   }
   else
   {
      if (-score-adjust2+lazy0 < alpha && board.getMaterial(Black).queen_count() < 2 )
	     alpha_cut++;
      if (-score+adjust1-lazy0 > beta && board.getMaterial(White).queen_count() < 2)
    	 beta_cut++;
   }
#else
   if (board.Side() == White) {
      if (lazy0-adjust2+score < alpha && board.getMaterial(White).queen_count() < 2)
    	 return score;
      if (score+adjust1-lazy0 > beta && board.getMaterial(Black).queen_count() < 2)
    	 return score;
   }
   else
   {
      if (-score-adjust2+lazy0 < alpha && board.getMaterial(Black).queen_count() < 2 )
	     return -score;
      if (-score+adjust1-lazy0 > beta && board.getMaterial(White).queen_count() < 2)
    	 return -score;
   }
#endif
#endif
      if (!kp_hash_entry->cover_valid) {
          calc_cover(board,board.KingPos(White),White,kp_hash_entry->white_cover,
                    kp_hash_entry->white_pawn_proximity);
          calc_cover(board,board.KingPos(Black),Black,kp_hash_entry->black_cover,
                    kp_hash_entry->black_pawn_proximity);
         kp_hash_entry->cover_valid = 1;
      } 
      score += king_safety(board,kp_hash_entry);
   }

   // We have evaluated the big parts of the total score. See
   // if we can exit now
#define LAZY1 133

   int lazy = LAZY1;
#ifdef LAZY_DEBUG
   int lazy_score1 = score;
#endif

#ifdef USE_LAZY_EVAL
    if (board.Side() == White) {
#if !defined(LAZY_DEBUG)
       if (lazy+score < alpha && board.getMaterial(White).queen_count() < 2)
    	 return score;
       if (score-lazy > beta) return score;
#endif
    }
    else
    {
#if !defined(LAZY_DEBUG)
       if (lazy-score < alpha && board.getMaterial(Black).queen_count() < 2 )
	     return -score;
       if (-score-lazy > beta) return -score;
#endif
   }
#endif
   
#ifdef USE_LAZY_EVAL   
   #define LAZY2 110
   int lazy2 = LAZY2;
#ifdef LAZY_DEBUG
   int lazy_score2=score;
#endif
//   if (!endgame)
   {
      if (board.Side() == White)
      {
         if (score + lazy2 < alpha)
         {
#ifdef _TRACE
            cout << "lazy cutoff, score =" << score << " alpha=" << alpha
            << endl;
#endif
#ifndef LAZY_DEBUG
            return score;
#endif            
         }
         if (score - lazy2 > beta)
         {
#ifdef _TRACE
            cout << "lazy cutoff, score =" << score << " beta=" << beta
            << endl;
            if (Util::Abs(score) > 250)
               cout << board << endl;
#endif
#ifndef LAZY_DEBUG
            return score;
#endif
         }
      } 
      else
      {
         if (-score + lazy2 < alpha)
         {
#ifdef _TRACE
            cout << "lazy cutoff, score =" << score << " alpha=" << alpha
            << endl;
#endif
#ifndef LAZY_DEBUG
            return -score;
#endif
         }
         if (-score - lazy2 > beta)
         {
#ifdef _TRACE
            cout << "lazy cutoff, score =" << score << " beta=" << beta
            << endl;
            if (Util::Abs(score) > 250)
               cout << board << endl;
#endif
#ifndef LAZY_DEBUG
            return -score;
#endif
         }
      }
   }
#endif

   Bitmap knights(board.knight_bits[White]);
   Square sq;
   for (;(sq = knights.first_one()) != InvalidSquare;knights.clear(sq))
   {
      int center = KnightScoresW[sq];
      if (w_material_level < 6) center /= 2;
      score += center;
      {
         // Compute knight mobility.  We award a bonus for
         // each square to which the knight can move, that
         // is not occupied or attacked by an enemy pawn.
         Bitmap ka(Bearing::knight_attacks[sq]);
         ka.andNot(board.all_occupied);
         ka.andNot(hash_entry_w->opponent_pawn_attacks);
         score += KNIGHT_MOBILITY[ka.bit_count()];
      }
      // evaluate a knight outpost
      if (OutpostScoresW[sq] &&
          Bitmap::And(outpostW[sq],board.pawn_bits[Black]).is_clear())
      {
          int outpost = OutpostScoresW[sq]+KNIGHT_OUTPOST;
          if (hash_entry_b->opponent_pawn_attacks.is_set(sq))
          {
             outpost += OutpostScoresW[sq]+KNIGHT_OUTPOST; // defended by pawns
          }
          if (w_material_level >= 6)
            score += outpost;
          else
            score += outpost/2;
      }
   }
   knights = board.knight_bits[Black];
   for (;(sq = knights.first_one()) != InvalidSquare;knights.clear(sq))
   {
      int center = KnightScoresB[sq];
      if (b_material_level < 6) center /= 2;
      score -= center;
      {
         // Compute knight mobility.  We award a bonus for
         // each square to which the knight can move, that
         // is not occupied or attacked by an enemy pawn.
         Bitmap ka(Bearing::knight_attacks[sq]);
         ka.andNot(board.all_occupied);
         ka.andNot(hash_entry_b->opponent_pawn_attacks);
         score -= KNIGHT_MOBILITY[ka.bit_count()];
      }
      // evaluate a knight outpost
      if (OutpostScoresB[sq] &&
          Bitmap::And(outpostB[sq],board.pawn_bits[White]).is_clear())
      {
          int outpost = OutpostScoresB[sq]+KNIGHT_OUTPOST;
          if (hash_entry_w->opponent_pawn_attacks.is_set(sq))
          {
             outpost += OutpostScoresB[sq]+KNIGHT_OUTPOST; // defended by pawns
          }
          if (b_material_level >= 6)
            score -= outpost;
          else
            score -= outpost/2;
      }
   }
   Bitmap queens(board.queen_bits[White]);
   if (!queens.is_clear())
   {
      Square sq = queens.first_one();

      if (FileOpen(File(sq),White) &&
          FileOpen(File(sq),Black))
         score += QUEEN_ON_OPEN_FILE;
      if (Rank(sq,White) == 2 && (Rank(board.KingPos(Black),Black)==1 ||
				  hash_entry_b->pawns_on_2nd_rank))
      {
         score += QUEEN_ON_7TH_RANK;
      }
      if (Rank(sq,White) > 2 && w_material_level > 6)
      {
         Bitmap b(board.knight_bits[White]);
         b.Or(board.bishop_bits[White]);
         b.And(Bearing::rank_mask[7]);
         int piecesBack = b.bit_count();
         if (piecesBack > 0)
         {
            if (board.castled(White))
               score += (QUEEN_OUT*(piecesBack+w_material_level/2))/10;
            else
               score += 3*(QUEEN_OUT*(piecesBack+w_material_level/2))/20;
	 }
      }
      score += QueenCenterScores[(int)sq];
   }
   queens = board.queen_bits[Black];
   if (!queens.is_clear())
   {
      Square sq = queens.first_one();
      if (FileOpen(File(sq),White) &&
          FileOpen(File(sq),Black))
         score -= QUEEN_ON_OPEN_FILE;

      if (Rank(sq,Black) == 2 && (Rank(board.KingPos(White),White)==1 ||
				  hash_entry_w->pawns_on_2nd_rank))
      {
         score -= QUEEN_ON_7TH_RANK;
      }
      if (Rank(sq,Black) > 2 && b_material_level > 6)
      {
         Bitmap b(board.knight_bits[Black]);
         b.Or(board.bishop_bits[Black]);
         b.And(Bearing::rank_mask[0]);
         int piecesBack = b.bit_count();
         if (piecesBack > 0)
         {
            if (board.castled(Black))
               score -= (QUEEN_OUT*(piecesBack+b_material_level/2))/10;
            else
               score -= 3*(QUEEN_OUT*(piecesBack+b_material_level/2))/20;
         }
      }
      score -= QueenCenterScores[(int)sq];
   }

   Bitmap rooks = board.rook_bits[White];
   for (;(sq = rooks.first_one()) != InvalidSquare;rooks.clear(sq))
   {
      const int file = File(sq);
      int mobl = 0;
      Square sq2 = sq;
      while (File(sq2--)!=1) {
          if (board[sq2] != EmptyPiece) break;
          if (!hash_entry_w->opponent_pawn_attacks.is_set(sq2)) {
            mobl++; break;
          }
      }
      sq2 = sq;
      while (File(sq2++)!=8) {
          if (board[sq2] != EmptyPiece) break;
          if (!hash_entry_w->opponent_pawn_attacks.is_set(sq2)) {
            mobl++; break;
          }
      }
      if (mobl==0) {
          score += LOW_ROOK_MOBILITY;
#ifdef EVAL_DEBUG
          cout << "low rook mobility(w): " << LOW_ROOK_MOBILITY << endl;
#endif
      }
      if (b_material_level >= 4 && FileOpen(file,White) &&
         (FileOpen(file,Black) || 
          (hash_entry_b->weak_files & (1<<(file-1)))))
      {
         score += ROOK_ON_OPEN_FILE;

         if (!Bitmap::And(board.rook_bits[White],Bearing::file_mask[file-1]).is_clear())
         {
            if (!Bitmap::And(FileAttacksUp(board,sq),board.rooksqueens[White]).is_clear())
               score += DOUBLED_ROOKS;
         }
      }
   }

   rooks = board.rook_bits[Black];
   for (;(sq = rooks.first_one()) != InvalidSquare;rooks.clear(sq))
   {
      const int file = File(sq);
      int mobl = 0;
      Square sq2 = sq;
      while (File(sq2--)!=1) {
          if (board[sq2] != EmptyPiece) break;
          if (!hash_entry_b->opponent_pawn_attacks.is_set(sq2)) {
            mobl++; break; 
          }
      }
      sq2 = sq;
      while (File(sq2++)!=8) {
          if (board[sq2] != EmptyPiece) break;
          if (!hash_entry_b->opponent_pawn_attacks.is_set(sq2)) {
             mobl++; break;
          }
      }
      if (mobl==0) {
          score -= LOW_ROOK_MOBILITY;
#ifdef EVAL_DEBUG
          cout << "low rook mobility(b): " << -LOW_ROOK_MOBILITY << endl;
#endif
      }
      if (w_material_level >= 4 && FileOpen(file,Black) &&
          (FileOpen(file,White) || 
           (hash_entry_w->weak_files & (1<<(file-1)))))
      {
         score -= ROOK_ON_OPEN_FILE;

         if (!Bitmap::And(board.rook_bits[Black],Bearing::file_mask[file-1]).is_clear())
         {
            if (!Bitmap::And(FileAttacksDown(board,sq),board.rooksqueens[Black]).is_clear())
               score -= DOUBLED_ROOKS;
         }
      }
   }

#ifdef USE_LAZY_EVAL   
#define LAZY3 62
#ifdef LAZY_DEBUG
   int lazy_score3=score;
#endif
//   if (!endgame)
   {
      if (board.Side() == White)
      {
         if (score + LAZY3 < alpha)
         {
#ifdef _TRACE
            cout << "lazy cutoff, score =" << score << " alpha=" << alpha
            << endl;
#endif
#ifndef LAZY_DEBUG
            return score;
#endif            
         }
         if (score - LAZY3 > beta)
         {
#ifdef _TRACE
            cout << "lazy cutoff, score =" << score << " beta=" << beta
            << endl;
            if (Util::Abs(score) > 250)
               cout << board << endl;
#endif
#ifndef LAZY_DEBUG
            return score;
#endif
         }
      } 
      else
      {
         if (-score + LAZY3 < alpha)
         {
#ifdef _TRACE
            cout << "lazy cutoff, score =" << score << " alpha=" << alpha
            << endl;
#endif
#ifndef LAZY_DEBUG
            return -score;
#endif
         }
         if (-score - LAZY3 > beta)
         {
#ifdef _TRACE
            cout << "lazy cutoff, score =" << score << " beta=" << beta
            << endl;
            if (Util::Abs(score) > 250)
               cout << board << endl;
#endif
#ifndef LAZY_DEBUG
            return -score;
#endif
         }
      }
   }
#endif


   int haveWbishop = 0 /*FALSE*/;
   int haveBbishop = 0 /*FALSE*/;
   if (!board.bishop_bits[White].is_clear())
   {
      Bitmap bishops(board.bishop_bits[White]);
      for (;(sq = bishops.first_one()) != InvalidSquare;bishops.clear(sq))
      {
         if (SquareColor(sq) == White)
           haveWbishop = 1 /*TRUE*/;
         else
           haveBbishop = 1 /*TRUE*/;
         if (OutpostScoresW[sq] &&
              Bitmap::And(outpostW[sq],board.pawn_bits[Black]).is_clear())
         {
            score += OutpostScoresW[sq];
            if (!Bitmap::And(Bearing::pawn_attacks[sq][White],
                board.pawn_bits[White]).is_clear())
            {
               score += OutpostScoresW[sq]; // defended by pawns
            }
         }
         if (!endgame)
         {
            score += BishopPositionW[sq];
         }
         score += BishopCenterScores[sq];
         int mobility = 0;
         int i = sq; int j = 0;
         while (File(i) != 8 && j < 2)
         {
            i -= 7;
            if (i < 0 ) break;
            if (board[i] != EmptyPiece)
            {
               break;
            }
            ++j;
         }
         mobility += j;
         i = sq; j = 0;
         while (File(i) != 1 && j < 2)
         {
            i -= 9;
            if (i < 0) break;
            if (board[i] != EmptyPiece)
            {
               break;
            }
            ++j;
         }
         mobility += j;
         if (File(sq) == 1 && sq+9<64 && board[sq+9] != EmptyPiece)
            mobility -= 1;
         if (File(sq) == 8 && sq+7<64 && board[sq+7] != EmptyPiece)
            mobility -= 1;
         score += BISHOP_MOBILITY[mobility+2];
      } 
   }
   if (haveWbishop && haveBbishop)
      score += TWO_BISHOPS*(16-b_material_level)/16;
   else if (haveWbishop && !haveBbishop)
   {
      int whitePawns = hash_entry_w->w_square_pawns;
      int blackPawns = hash_entry_w->b_square_pawns;
      if ((whitePawns + blackPawns > 4) && (whitePawns > blackPawns))
         score += BAD_BISHOP;
   } 
   else if (haveBbishop && !haveWbishop)
   {
      int whitePawns = hash_entry_w->w_square_pawns;
      int blackPawns = hash_entry_w->b_square_pawns;
      if ((whitePawns + blackPawns > 4) && (blackPawns > whitePawns))
         score += BAD_BISHOP;
   }
   haveWbishop = haveBbishop = 0 /*FALSE*/;

   if (!board.bishop_bits[Black].is_clear())
   {
      Bitmap bishops(board.bishop_bits[Black]);
      Square sq;
      for (;(sq = bishops.first_one()) != InvalidSquare;bishops.clear(sq))
      {
         if (SquareColor(sq) == White)
           haveWbishop = 1 /*TRUE*/;
         else
           haveBbishop = 1 /*TRUE*/;
         if (OutpostScoresB[sq] &&
              Bitmap::And(outpostB[sq],board.pawn_bits[White]).is_clear())
         {
            score -= OutpostScoresB[sq];
            if (!Bitmap::And(Bearing::pawn_attacks[sq][Black],
                board.pawn_bits[Black]).is_clear())
            {
               score -= OutpostScoresB[sq]; // defended by pawns
            }
         }
         if (!endgame)
         {
            score -= BishopPositionB[sq];
         }
         score -= BishopCenterScores[sq];
         int mobility = 0;
         int i = sq; int j = 0;
         while (File(i) != 8 && j < 2)
         {
            i += 9;
            if (i > 63) break;
            if (board[i] != EmptyPiece)
            {
               break;
            }
            ++j;
         }
         mobility += j;
         i = sq; j = 0;
         while (File(i) != 1 && j < 2)
         {
            i += 7;
            if (i>63) break;
            if (board[i] != EmptyPiece)
            {
               break;
            }
            ++j;
         }
         mobility += j;
         if (File(sq) == 1 && (int)sq-7>0 && board[sq-7] != EmptyPiece)
            mobility -= 1;
         if (File(sq) == 8 && (int)sq-9>0 && board[sq-9] != EmptyPiece)
            mobility -= 1;
         score -= BISHOP_MOBILITY[mobility+2];
      } 
   }

   if (haveWbishop && haveBbishop)
      score -= TWO_BISHOPS*(16-w_material_level)/16;
   if (haveWbishop && !haveBbishop)
   {
      int whitePawns = hash_entry_b->w_square_pawns;
      int blackPawns = hash_entry_b->b_square_pawns;
      if ((whitePawns + blackPawns > 4) && (whitePawns > blackPawns))
         score -= BAD_BISHOP;
   } 
   else if (haveBbishop && !haveWbishop)
   {
      int whitePawns = hash_entry_b->w_square_pawns;
      int blackPawns = hash_entry_b->b_square_pawns;
      if ((whitePawns + blackPawns > 4) && (blackPawns > whitePawns))
         score -= BAD_BISHOP;
   }
   if (!endgame)
   {
      // Don't block center pawns with pieces
      if (board[D2] == whitePawn)
      {
         Piece p2 = board[D3];
         if (TypeOfPiece(p2) != EmptyPiece && 
             TypeOfPiece(p2) != Pawn && 
             PieceColor(p2) == White)
            score += CENTER_PAWN_BLOCK;
      }
      if (board[E2] == whitePawn)
      {
         Piece p2 = board[E3];
         if (TypeOfPiece(p2) != EmptyPiece && 
             TypeOfPiece(p2) != Pawn &&
             PieceColor(p2) == White)
            score += CENTER_PAWN_BLOCK;
      }
      if (board[D7] == blackPawn)
      {
         Piece p2 = board[D6];
         if (TypeOfPiece(p2) != EmptyPiece && 
             TypeOfPiece(p2) != Pawn &&
             PieceColor(p2) == Black)
            score -= CENTER_PAWN_BLOCK;
      }
      if (board[E7] == blackPawn)
      {
         Piece p2 = board[E6];
         if (TypeOfPiece(p2) != EmptyPiece && 
             TypeOfPiece(p2) != Pawn &&
             PieceColor(p2) == Black)
            score -= CENTER_PAWN_BLOCK;
      }
   }

#ifdef LAZY_DEBUG
   if (board.Side() == White) {
      if (alpha_cut)
         if (score >= alpha)
    	   cout << board << endl;
      if (beta_cut)
          if (score <= beta)
             cout << board << endl;
   }
   else
   {
      if (alpha_cut)
         if (-score >= alpha)
    	   cout << board << endl;
      if (beta_cut)
          if (-score <= beta)
             cout << board << endl;
   }

   if (Util::Abs(lazy_score1 -score) > LAZY1)
      cout << board << 
       " difference1 = " << Util::Abs(lazy_score1 -score) << endl;
   if (Util::Abs(lazy_score2 -score) > lazy2)
      cout << board << 
       " difference2 = " << Util::Abs(lazy_score2 -score) << endl;
   if (Util::Abs(lazy_score3 -score) > LAZY3)
      cout << board << 
       " difference3 = " << Util::Abs(lazy_score3 -score) << endl;
#endif
   if (board.Side() == White)
      return score;
   else
      return -score;
}


int Scoring::material_score( const Board &board )
// returns a material score
{
   return MaterialScore(board,board.Side());
}

int Scoring::material_score( const Board &board, Move m )
{
   Material & oppmat = (Material&)board.getMaterial(board.OppositeSide());
   Material & ourmat = (Material&)board.getMaterial(board.Side());
   // Adjust for trade bonus
   oppmat.remove_piece(Capture(m));
   if (TypeOfMove(m) == Promotion)
   {
     ourmat.remove_piece(Pawn);
     ourmat.add_piece(PromoteTo(m));
   }
   int mat_score = MaterialScore(board,board.Side());
   oppmat.add_piece(Capture(m));
   if (TypeOfMove(m) == Promotion)
   {
     ourmat.add_piece(Pawn);
     ourmat.remove_piece(PromoteTo(m));
   }
   return mat_score;
}

int Scoring::repetition_draw( const Board &board )
{
   return board.rep_count() >= 2;
}

int Scoring::material_draw( const Board &board)
{
   // check for insufficient material
   const Material &mat1 = board.getMaterial(White);
   const Material &mat2 = board.getMaterial(Black);
   if (mat1.pawn_count() || mat2.pawn_count())
   {
      return 0 /*FALSE*/;
   }
   if ((mat1.value() <= KING_VALUE + BISHOP_VALUE) &&
       (mat2.value() <= KING_VALUE + BISHOP_VALUE))
   {
      if (mat1.king_only() || mat2.king_only())
         // K vs K, or K vs KN, or K vs KB
         return 1 /*TRUE*/;
      else if (mat1.knight_count() || mat2.knight_count())
      {
         // KN vs KN 
         return 0 /*FALSE*/;
      } else
      {
         Bitmap wbishops(board.bishop_bits[White]);
         Bitmap bbishops(board.bishop_bits[Black]);
         // drawn if same-color bishops
         if
             (!(Bitmap::And(wbishops,Bearing::black_squares)).is_clear())
         {
            return !(Bitmap::And(bbishops,Bearing::black_squares)).is_clear();
         } else
         {
            return !(Bitmap::And(bbishops,Bearing::white_squares)).is_clear();
         }
      }
   }
   return 0 /*FALSE*/;
}

const char* Scoring::print_score( const int score ) {
  static char buf[20];
  if (score <= -Constants::BIG+100) {
      int movesToMate = (Constants::BIG+score+1)/2;
      sprintf(buf,"-Mate%d",movesToMate);
  }
  else if (score >= Constants::BIG-100) {
      int movesToMate = (Constants::BIG-score+1)/2;
      sprintf(buf,"+Mate%d",movesToMate);
  }
  else
    sprintf(buf,"%+6.2f",(1.0F*score)/64.0F);
  return buf;
}
